from __future__ import annotations
import abc
import gc
import itertools
from copy import deepcopy
from typing import Union, Dict, List, Optional, TYPE_CHECKING

from abc import ABC

if TYPE_CHECKING:
    from clayrs.content_analyzer import Ratings
    from clayrs.recsys.graphs.graph import FullDiGraph
    from clayrs.recsys.content_based_algorithm.content_based_algorithm import ContentBasedAlgorithm
    from clayrs.recsys.graph_based_algorithm.graph_based_algorithm import GraphBasedAlgorithm
    from clayrs.recsys.methodology import Methodology

from clayrs.content_analyzer.ratings_manager.ratings import Rank, Prediction
from clayrs.recsys.methodology import TestRatingsMethodology
from clayrs.recsys.content_based_algorithm.exceptions import UserSkipAlgFit, NotFittedAlg
from clayrs.utils.const import logger
from clayrs.utils.context_managers import get_iterator_parallel


class RecSys(ABC):
    """
    Abstract class for a Recommender System

    There exists various type of recommender systems, content-based, graph-based, etc. so extend this class
    if another type must be implemented into the framework.

    Every recommender system has an algorithm which is used to compute rank/score predictions

    Args:
        algorithm: The algorithm used to compute rank/score prediction
    """

    def __init__(self, algorithm: Union[ContentBasedAlgorithm, GraphBasedAlgorithm]):
        self.__alg = algorithm

        self._yaml_report: Optional[Dict] = None

    @property
    def algorithm(self):
        return self.__alg

    @abc.abstractmethod
    def rank(self, test_set: Ratings, n_recs: int = 10) -> Rank:
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, test_set: Ratings) -> Prediction:
        raise NotImplementedError


class ContentBasedRS(RecSys):
    """
    Class for recommender systems which use the items' content in order to make predictions,
    some algorithms may also use users' content, so it's an optional parameter.

    Every CBRS differ from each other based the algorithm used.

    Examples:

        In case you perform a splitting of the dataset which returns a single train and test set (e.g. HoldOut
        technique):

        ```python title="Single split train"
        from clayrs import recsys as rs
        from clayrs import content_analyzer as ca

        original_rat = ca.Ratings(ca.CSVFile(ratings_path))

        [train], [test] = rs.HoldOutPartitioning().split_all(original_rat)

        alg = rs.CentroidVector()  # any cb algorithm

        cbrs = rs.ContentBasedRS(alg, train, items_path)

        rank = cbrs.fit_rank(test, n_recs=10)
        ```

        In case you perform a splitting of the dataset which returns a multiple train and test sets (KFold technique):

        ```python title="Multiple split train"
        from clayrs import recsys as rs
        from clayrs import content_analyzer as ca

        original_rat = ca.Ratings(ca.CSVFile(ratings_path))

        train_list, test_list = rs.KFoldPartitioning(n_splits=5).split_all(original_rat)

        alg = rs.CentroidVector()  # any cb algorithm

        for train_set, test_set in zip(train_list, test_list):

            cbrs = rs.ContentBasedRS(alg, train_set, items_path)
            rank_to_append = cbrs.fit_rank(test_set)

            result_list.append(rank_to_append)
        ```

        `result_list` will contain recommendation lists for each split

    Args:
        algorithm: the content based algorithm that will be used in order to
            rank or make score prediction
        train_set: a Ratings object containing interactions between users and items
        items_directory: the path of the items serialized by the Content Analyzer
        users_directory: the path of the users serialized by the Content Analyzer
    """

    def __init__(self,
                 algorithm: ContentBasedAlgorithm,
                 train_set: Ratings,
                 items_directory: str,
                 users_directory: str = None):

        super().__init__(algorithm)
        self.__train_set = train_set
        self.__items_directory = items_directory
        self.__users_directory = users_directory
        self._user_fit_dic = {}

    @property
    def algorithm(self):
        """
        The content based algorithm chosen
        """
        alg: ContentBasedAlgorithm = super().algorithm
        return alg

    @property
    def train_set(self):
        """
        The train set of the Content Based RecSys
        """
        return self.__train_set

    @property
    def items_directory(self):
        """
        Path of the serialized items by the Content Analyzer
        """
        return self.__items_directory

    @property
    def users_directory(self):
        """
        Path of the serialized users by the Content Analyzer
        """
        return self.__users_directory

    def fit(self, num_cpus: int = 0):
        """
        Method which will fit the algorithm chosen for each user in the train set passed in the constructor

        If the algorithm can't be fit for some users, a warning message is printed
        """
        def compute_single_fit(user_id):
            user_train = self.train_set.get_user_interactions(user_id)
            user_alg = deepcopy(self.algorithm)

            try:
                user_alg.process_rated(user_train, loaded_items_interface)
                user_alg.fit()
            except UserSkipAlgFit as e:
                warning_message = str(e) + f"\nNo algorithm will be fitted for the user {user_id}"
                logger.warning(warning_message)
                user_alg = None

            return user_id, user_alg

        items_to_load = set(self.train_set.item_id_column)
        all_users = set(self.train_set.user_id_column)
        loaded_items_interface = self.algorithm._load_available_contents(self.items_directory, items_to_load)

        with get_iterator_parallel(num_cpus,
                                   compute_single_fit, all_users,
                                   progress_bar=True, total=len(all_users)) as pbar:

            pbar.set_description("Fitting algorithm")

            for user_id, fitted_user_alg in pbar:
                self._user_fit_dic[user_id] = fitted_user_alg

        # we force the garbage collector after freeing loaded items
        del loaded_items_interface
        gc.collect()

        return self

    def rank(self, test_set: Ratings, n_recs: int = 10, user_id_list: List = None,
             methodology: Union[Methodology, None] = TestRatingsMethodology(),
             num_cpus: int = 1) -> Rank:

        """
        Method used to calculate ranking for all users in test set or all users in `user_id_list` parameter.
        You must first call the `fit()` method before you can compute the ranking.

        If the `n_recs` is specified, then the rank will contain the top-n items for the users.
        Otherwise, the rank will contain all unrated items of the particular users

        Via the `methodology` parameter you can perform different candidate item selection. By default, the
        `TestRatingsMethodology()` is used: so, for each user, items in its test set only will be ranked

        If the algorithm was not fit for some users, they will be skipped and a warning is printed

        Args:
            test_set: Ratings object which represents the ground truth of the split considered
            n_recs: Number of the top items that will be present in the ranking of each user.
                If `None` all candidate items will be returned for the user
            user_id_list: List of users for which you want to compute the ranking. If None, the ranking will be computed
                for all users of the `test_set`
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected.

        Raises:
            NotFittedAlg: Exception raised when this method is called without first calling the `fit` method

        Returns:
            Rank object containing recommendation lists for all users of the test set or for all users in `user_id_list`

        """
        def compute_single_rank(user_id):
            user_id = str(user_id)
            user_train = self.train_set.get_user_interactions(user_id)

            filter_list = None
            if methodology is not None:
                filter_list = set(methodology.filter_single(user_id, self.train_set, test_set))

            user_fitted_alg = self._user_fit_dic.get(user_id)
            if user_fitted_alg is not None:
                user_rank = user_fitted_alg.rank(user_train, loaded_items_interface,
                                                 n_recs, filter_list=filter_list)
            else:
                user_rank = []
                logger.warning(f"No algorithm fitted for user {user_id}! It will be skipped")

            return user_id, user_rank

        if len(self._user_fit_dic) == 0:
            raise NotFittedAlg("Algorithm not fit! You must call the fit() method first, or fit_rank().")

        all_users = set(test_set.user_id_column)
        if user_id_list is not None:
            all_users = set(user_id_list)

        loaded_items_interface = self.algorithm._load_available_contents(self.items_directory, set())

        rank = []

        logger.info("Don't worry if it looks stuck at first")
        logger.info("First iterations will stabilize the estimated remaining time")

        with get_iterator_parallel(num_cpus,
                                   compute_single_rank, all_users,
                                   progress_bar=True, total=len(all_users)) as pbar:

            pbar.set_description(f"Loading first items from memory...")
            for user_id, user_rank in pbar:
                pbar.set_description(f"Computing rank for user {user_id}")
                rank.append(user_rank)

        rank = itertools.chain.from_iterable(rank)
        rank = Rank.from_list(rank)

        # we force the garbage collector after freeing loaded items
        del loaded_items_interface
        gc.collect()

        self._yaml_report = {'mode': 'rank', 'n_recs': repr(n_recs), 'methodology': repr(methodology)}

        return rank

    def predict(self, test_set: Ratings, user_id_list: List = None,
                methodology: Union[Methodology, None] = TestRatingsMethodology(),
                num_cpus: int = 1) -> Prediction:
        """
        Method used to calculate score predictions for all users in test set or all users in `user_id_list` parameter.
        You must first call the `fit()` method before you can compute score predictions.

        **BE CAREFUL**: not all algorithms are able to perform *score prediction*

        Via the `methodology` parameter you can perform different candidate item selection. By default, the
        `TestRatingsMethodology()` is used: so, for each user, items in its test set only will be considered for score
        prediction

        If the algorithm was not fit for some users, they will be skipped and a warning is printed

        Args:
            test_set: Ratings object which represents the ground truth of the split considered
            user_id_list: List of users for which you want to compute score prediction. If None, the ranking
                will be computed for all users of the `test_set`
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected.

        Raises:
            NotFittedAlg: Exception raised when this method is called without first calling the `fit` method

        Returns:
            Prediction object containing score prediction lists for all users of the test set or for all users in
                `user_id_list`

        """
        def compute_single_predict(user_id):
            user_id = str(user_id)
            user_train = self.train_set.get_user_interactions(user_id)

            filter_list = None
            if methodology is not None:
                filter_list = set(methodology.filter_single(user_id, self.train_set, test_set))

            user_fitted_alg = self._user_fit_dic.get(user_id)
            if user_fitted_alg is not None:
                user_pred = user_fitted_alg.predict(user_train, loaded_items_interface, filter_list=filter_list)
            else:
                user_pred = []
                logger.warning(f"No algorithm fitted for user {user_id}! It will be skipped")

            return user_id, user_pred

        if len(self._user_fit_dic) == 0:
            raise NotFittedAlg("Algorithm not fit! You must call the fit() method first, or fit_rank().")

        all_users = set(test_set.user_id_column)
        if user_id_list is not None:
            all_users = set(user_id_list)

        loaded_items_interface = self.algorithm._load_available_contents(self.items_directory, set())

        pred = []

        logger.info("Don't worry if it looks stuck at first")
        logger.info("First iterations will stabilize the estimated remaining time")

        with get_iterator_parallel(num_cpus,
                                   compute_single_predict, all_users,
                                   progress_bar=True, total=len(all_users)) as pbar:

            pbar.set_description(f"Loading first items from memory...")
            for user_id, user_pred in pbar:
                pbar.set_description(f"Computing score prediction for user {user_id}")
                pred.append(user_pred)

        pred = itertools.chain.from_iterable(pred)
        pred = Prediction.from_list(pred)

        # we force the garbage collector after freeing loaded items
        del loaded_items_interface
        gc.collect()

        self._yaml_report = {'mode': 'score_prediction', 'methodology': repr(methodology)}

        return pred

    def fit_predict(self, test_set: Ratings, user_id_list: List = None,
                    methodology: Union[Methodology, None] = TestRatingsMethodology(),
                    save_fit: bool = False, num_cpus: int = 1) -> Prediction:
        """
        Method used to both fit and calculate score prediction for all users in test set or all users in `user_id_list`
        parameter.
        The Recommender System will first be fit for each user in the train set passed in the constructor.
        If the algorithm can't be fit for some users, a warning message is printed

        **BE CAREFUL**: not all algorithms are able to perform *score prediction*

        Via the `methodology` parameter you can perform different candidate item selection. By default, the
        `TestRatingsMethodology()` is used: so, for each user, items in its test set only will be considered for score
        prediction

        If the algorithm was not fit for some users, they will be skipped and a warning is printed

        With the `save_fit` parameter you can decide if you want that you recommender system remains *fit* even after
        the complete execution of this method, in case you want to compute ranking/score prediction with other
        methodologies, or with a different `n_recs` parameter. Be mindful since it can be memory-expensive,
        thus by default this behaviour is disabled

        Args:
            test_set: Ratings object which represents the ground truth of the split considered
            user_id_list: List of users for which you want to compute score prediction. If None, the ranking
                will be computed for all users of the `test_set`
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            save_fit: Boolean value which let you choose if the Recommender System should remain fit even after the
                complete execution of this method. Default is False
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected.

        Returns:
            Rank object containing recommendation lists for all users of the test set or for all users in `user_id_list`
        """
        def compute_single_fit_predict(user_id):
            user_train = self.train_set.get_user_interactions(user_id)

            alg = self.algorithm
            if save_fit:
                alg = deepcopy(alg)
                self._user_fit_dic[user_id] = alg

            try:
                alg.process_rated(user_train, loaded_items_interface)
                alg.fit()

            except UserSkipAlgFit as e:
                warning_message = str(e) + f"\nThe algorithm can't be fitted for the user {user_id}"
                logger.warning(warning_message)
                if save_fit:
                    self._user_fit_dic[user_id] = None
                return user_id, []

            filter_list = None
            if methodology is not None:
                filter_list = set(methodology.filter_single(user_id, self.train_set, test_set))

            user_pred = alg.predict(user_train, loaded_items_interface, filter_list=filter_list)

            return user_id, user_pred

        all_users = set(test_set.user_id_column)
        if user_id_list is not None:
            all_users = set(user_id_list)

        loaded_items_interface = self.algorithm._load_available_contents(self.items_directory, set())

        pred = []

        logger.info("Don't worry if it looks stuck at first")
        logger.info("First iterations will stabilize the estimated remaining time")

        with get_iterator_parallel(num_cpus,
                                   compute_single_fit_predict, all_users,
                                   progress_bar=True, total=len(all_users)) as pbar:

            pbar.set_description(f"Loading first items from memory...")
            for user_id, user_pred in pbar:
                pbar.set_description(f"Computing fit_predict for user {user_id}")
                pred.append(user_pred)

        pred = itertools.chain.from_iterable(pred)
        pred = Prediction.from_list(pred)

        # we force the garbage collector after freeing loaded items
        del loaded_items_interface
        gc.collect()

        self._yaml_report = {'mode': 'score_prediction', 'methodology': repr(methodology)}

        return pred

    def fit_rank(self, test_set: Ratings, n_recs: int = 10, user_id_list: List = None,
                 methodology: Union[Methodology, None] = TestRatingsMethodology(),
                 save_fit: bool = False, num_cpus: int = 1) -> Rank:
        """
        Method used to both fit and calculate ranking for all users in test set or all users in `user_id_list`
        parameter.
        The Recommender System will first be fit for each user in the train set passed in the constructor.
        If the algorithm can't be fit for some users, a warning message is printed

        If the `n_recs` is specified, then the rank will contain the top-n items for the users.
        Otherwise, the rank will contain all unrated items of the particular users

        Via the `methodology` parameter you can perform different candidate item selection. By default, the
        `TestRatingsMethodology()` is used: so, for each user, items in its test set only will be ranked

        If the algorithm was not fit for some users, they will be skipped and a warning is printed

        With the `save_fit` parameter you can decide if you want that you recommender system remains *fit* even after
        the complete execution of this method, in case you want to compute ranking with other methodologies, or
        with a different `n_recs` parameter. Be mindful since it can be memory-expensive, thus by default this behaviour
        is disabled

        Args:
            test_set: Ratings object which represents the ground truth of the split considered
            n_recs: Number of the top items that will be present in the ranking of each user.
                If `None` all candidate items will be returned for the user
            user_id_list: List of users for which you want to compute the ranking. If None, the ranking will be computed
                for all users of the `test_set`
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            save_fit: Boolean value which let you choose if the Recommender System should remain fit even after the
                complete execution of this method. Default is False
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected.

        Raises:
            NotFittedAlg: Exception raised when this method is called without first calling the `fit` method

        Returns:
            Rank object containing recommendation lists for all users of the test set or for all users in `user_id_list`
        """
        def compute_single_fit_rank(user_id):
            user_train = self.train_set.get_user_interactions(user_id)

            alg = self.algorithm
            if save_fit:
                alg = deepcopy(alg)
                self._user_fit_dic[user_id] = alg

            try:
                alg.process_rated(user_train, loaded_items_interface)
                alg.fit()

            except UserSkipAlgFit as e:
                warning_message = str(e) + f"\nThe algorithm can't be fitted for the user {user_id}"
                logger.warning(warning_message)
                if save_fit:
                    self._user_fit_dic[user_id] = None
                return user_id, []

            filter_list = None
            if methodology is not None:
                filter_list = set(methodology.filter_single(user_id, self.train_set, test_set))

            user_rank = alg.rank(user_train, loaded_items_interface,
                                 n_recs, filter_list=filter_list)

            return user_id, user_rank

        all_users = set(test_set.user_id_column)
        if user_id_list is not None:
            all_users = set(user_id_list)

        loaded_items_interface = self.algorithm._load_available_contents(self.items_directory, set())

        rank = []

        logger.info("Don't worry if it looks stuck at first")
        logger.info("First iterations will stabilize the estimated remaining time")

        with get_iterator_parallel(num_cpus,
                                   compute_single_fit_rank, all_users,
                                   progress_bar=True, total=len(all_users)) as pbar:

            pbar.set_description(f"Loading first items from memory...")
            for user_id, user_rank in pbar:
                pbar.set_description(f"Computing fit_rank for user {user_id}")
                rank.append(user_rank)

        rank = itertools.chain.from_iterable(rank)
        rank = Rank.from_list(rank)

        # we force the garbage collector after freeing loaded items
        del loaded_items_interface
        gc.collect()

        self._yaml_report = {'mode': 'rank', 'n_recs': repr(n_recs), 'methodology': repr(methodology)}

        return rank

    def __repr__(self):
        return f"ContentBasedRS(algorithm={self.algorithm}, train_set={self.train_set}, " \
               f"items_directory={self.items_directory}, users_directory={self.users_directory})"


class GraphBasedRS(RecSys):
    """
    Class for recommender systems which use a graph in order to make predictions

    Every GBRS differ from each other based the algorithm used.

    Examples:

        In case you perform a splitting of the dataset which returns a single train and test set (e.g. HoldOut
        technique):

        ```python title="Single split train"
        from clayrs import recsys as rs
        from clayrs import content_analyzer as ca

        original_rat = ca.Ratings(ca.CSVFile(ratings_path))

        [train], [test] = rs.HoldOutPartitioning().split_all(original_rat)

        alg = rs.NXPageRank()  # any gb algorithm

        graph = rs.NXBipartiteGraph(train)

        gbrs = rs.GraphBasedRS(alg, graph)

        rank = gbrs.rank(test, n_recs=10)
        ```

        In case you perform a splitting of the dataset which returns a multiple train and test sets (KFold technique):

        ```python title="Multiple split train"
        from clayrs import recsys as rs
        from clayrs import content_analyzer as ca

        original_rat = ca.Ratings(ca.CSVFile(ratings_path))

        train_list, test_list = rs.KFoldPartitioning(n_splits=5).split_all(original_rat)

        alg = rs.NXPageRank()  # any gb algorithm

        for train_set, test_set in zip(train_list, test_list):

            graph = rs.NXBipartiteGraph(train_set)
            gbrs = rs.GraphBasedRS(alg, graph)
            rank_to_append = gbrs.rank(test_set)

            result_list.append(rank_to_append)
        ```

        `result_list` will contain recommendation lists for each split

    Args:
        algorithm: the graph based algorithm that will be used in order to
            rank or make score prediction
        graph: a Graph object containing interactions
    """

    def __init__(self,
                 algorithm: GraphBasedAlgorithm,
                 graph: FullDiGraph):

        self.__graph = graph
        super().__init__(algorithm)

    @property
    def users(self):
        return self.__graph.user_nodes

    @property
    def graph(self):
        """
        The graph containing interactions
        """
        return self.__graph

    @property
    def algorithm(self):
        """
        The graph based algorithm chosen
        """
        alg: GraphBasedAlgorithm = super().algorithm
        return alg

    def predict(self, test_set: Ratings, user_id_list: List = None,
                methodology: Union[Methodology, None] = TestRatingsMethodology(),
                num_cpus: int = 1) -> Prediction:
        """
        Method used to calculate score predictions for all users in test set or all users in `user_id_list` parameter.

        **BE CAREFUL**: not all algorithms are able to perform *score prediction*

        Via the `methodology` parameter you can perform different candidate item selection. By default, the
        `TestRatingsMethodology()` is used: so for each user items in its test set only will be considered for score
        prediction

        If the algorithm was not fit for some users, they will be skipped and a warning is printed

        Args:
            test_set: Ratings object which represents the ground truth of the split considered
            user_id_list: List of users for which you want to compute score prediction. If None, the ranking
                will be computed for all users of the `test_set`
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected.

        Returns:
            Prediction object containing score prediction lists for all users of the test set or for all users in
                `user_id_list`

        """
        all_users = set(test_set.user_id_column)
        if user_id_list is not None:
            all_users = set(user_id_list)

        total_predict_list = self.algorithm.predict(all_users, self.graph, test_set, methodology, num_cpus)

        total_predict = Prediction.from_list(total_predict_list)

        self._yaml_report = {'graph': repr(self.graph), 'mode': 'score_prediction', 'methodology': repr(methodology)}

        return total_predict

    def rank(self, test_set: Ratings, n_recs: int = 10, user_id_list: List = None,
             methodology: Union[Methodology, None] = TestRatingsMethodology(),
             num_cpus: int = 1) -> Rank:
        """
        Method used to calculate ranking for all users in test set or all users in `user_id_list` parameter.
        You must first call the `fit()` method before you can compute the ranking.

        If the `n_recs` is specified, then the rank will contain the top-n items for the users.
        Otherwise, the rank will contain all unrated items of the particular users

        Via the `methodology` parameter you can perform different candidate item selection. By default, the
        `TestRatingsMethodology()` is used: so, for each user, items in its test set only will be ranked

        If the algorithm was not fit for some users, they will be skipped and a warning is printed

        Args:
            test_set: Ratings object which represents the ground truth of the split considered
            n_recs: Number of the top items that will be present in the ranking of each user.
                If `None` all candidate items will be returned for the user
            user_id_list: List of users for which you want to compute the ranking. If None, the ranking will be computed
                for all users of the `test_set`
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected.

        Raises:
            NotFittedAlg: Exception raised when this method is called without first calling the `fit` method

        Returns:
            Rank object containing recommendation lists for all users of the test set or for all users in `user_id_list`

        """
        all_users = set(test_set.user_id_column)
        if user_id_list is not None:
            all_users = set(user_id_list)

        total_rank_list = self.algorithm.rank(all_users, self.graph, test_set, n_recs, methodology, num_cpus)

        total_rank = Rank.from_list(total_rank_list)

        if len(total_rank) == 0:
            logger.warning("No items could be ranked for any users! Remember that items to rank must be present "
                           "in the graph.\n"
                           "Try changing methodology!")

        elif len(set(total_rank.user_id_column)) != len(all_users):
            logger.warning(f"No items could be ranked for users {all_users - set(total_rank.user_id_column)}\n"
                           f"No nodes to rank for them found in the graph. Try changing methodology! ")

        self._yaml_report = {'graph': repr(self.graph), 'mode': 'rank', 'n_recs': repr(n_recs),
                             'methodology': repr(methodology)}

        return total_rank

    def __repr__(self):
        return f"GraphBasedRS(algorithm={self.algorithm}, graph={self.graph})"



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.recsys.partitioning import Split


class Metric(ABC):
    """
    Abstract class that generalize metric concept

    Every metric may need different kind of "prediction": some (eg. NDCG, MRR, etc.) may need recommendation lists in
    which the recsys ranks every unseen item, some (eg. MAE, RMSE, etc.) may need a score prediction where the recsys
    must predict the rating that a user would give to an unseen item.
    So a Metric category (subclass of this class) must implement the "eval_fit_recsys(...)" specifying its needs,
    while every single metric (subclasses of the metric category class) must implement the "perform(...)" method
    specifying how to execute the metric computation
    """

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def perform(self, split: Split):
        raise NotImplementedError



from __future__ import annotations
import functools
import itertools
import os
from collections import defaultdict
from pathlib import Path
from typing import Dict, Union, List, Iterable, Iterator, TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from clayrs.content_analyzer.ratings_manager.score_processor import ScoreProcessor
    from clayrs.content_analyzer.raw_information_source import RawInformationSource

from clayrs.content_analyzer.exceptions import Handler_ScoreNotFloat
from clayrs.utils.context_managers import get_progbar
from clayrs.utils.save_content import get_valid_filename


class Interaction:
    """
    Class which models an interaction between a user and an item

    Each interaction has a score (a numeric value) and an optional timestamp

    Args:
        user_id: ID of the user
        item_id: ID of the item
        score: Numeric value of the interaction
        timestamp: Optional timestamp of the interaction
    """
    __slots__ = ('_user_id', '_item_id', '_score', '_timestamp')

    def __init__(self, user_id: str, item_id: str, score: float, timestamp: str = None):
        self._user_id = user_id
        self._item_id = item_id
        self._score = score
        self._timestamp = timestamp

    @property
    def user_id(self):
        """
        Getter for the `user_id` of the interaction
        """
        return self._user_id

    @property
    def item_id(self):
        """
        Getter for the `item_id` of the interaction
        """
        return self._item_id

    @property
    def score(self):
        """
        Getter for the `score` of the interaction
        """
        return self._score

    @property
    def timestamp(self):
        """
        Getter for the `timestamp` of the interaction. Could be `None`
        """
        return self._timestamp

    def __str__(self):
        if self.timestamp is not None:
            string = f"(user_id: {self.user_id}, item_id: {self.item_id}, score: {self.score}," \
                     f" timestamp: {self.timestamp})"
        else:
            string = f"(user_id: {self.user_id}, item_id: {self.item_id}, score: {self.score})"

        return string

    def __repr__(self):
        return f"Interaction(user_id={self.user_id}, item_id={self.item_id}, score={self.score}, " \
               f"timestamp={self.timestamp})"

    def __eq__(self, other):
        if isinstance(other, Interaction):
            timestamp_equal = (self.timestamp is None and other.timestamp is None) or \
                              (self.timestamp == other.timestamp)

            return self.user_id == other.user_id and self.item_id == other.item_id and \
                   self.score == other.score and timestamp_equal
        else:
            return False

    def __hash__(self):
        return hash(self.user_id + self.item_id + str(self.score) + str(self.timestamp))


class Ratings:
    """
    Class responsible of importing an interaction frame into the framework

    **If** the source file contains users, items and ratings in this order,
    no additional parameters are needed, **otherwise** the mapping must be explicitly specified using:

    * **'user_id'** column,
    * **'item_id'** column,
    * **'score'** column

    The *score* column can also be processed: in case you would like to consider as score the sentiment of a textual
    review, or maybe normalizing all scores in $[0, 1]$ range. Check the example below for more

    Examples:

        ```title="CSV raw source"
        user_id,item_id,rating,timestamp,review
        u1,i1,4,00112,good movie
        u2,i1,3,00113,an average movie
        u2,i32,2,00114,a bad movie
        ```

        As you can see the user id column, item id column and score column are the first three column and are already in
        sequential order, so no additional parameter is required to the Ratings class:

        >>> import clayrs.content_analyzer as ca
        >>> ratings_raw_source = ca.CSVFile('ratings.csv')
        >>> # add timestamp='timestamp' to the following if
        >>> # you want to load also the timestamp
        >>> ratings = ca.Ratings(ratings_raw_source)

        In case columns in the raw source are not in the above order you must specify an appropriate mapping via
        positional index (useful in case your raw source doesn't have a header) or via column ids:

        >>> # (mapping by index) EQUIVALENT:
        >>> ratings = ca.Ratings(
        >>> ca.CSVFile('ratings.csv'),
        >>> user_id_column=0,  # (1)
        >>> item_id_column=1,  # (2)
        >>> score_column=2  # (3)
        >>> )

        1. First column of raw source is the column containing all user ids
        2. Second column of raw source is the column containing all item ids
        3. Third column of raw source is the column containing all the scores

        >>> # (mapping by column name) EQUIVALENT:
        >>> ratings = ca.Ratings(
        >>> ca.CSVFile('ratings.csv'),
        >>> user_id_column='user_id',  # (1)
        >>> item_id_column='item_id',  # (2)
        >>> score_column='rating'  # (3)
        >>> )

        1. The column with id 'user_id' of raw source is the column containing all user ids
        2. The column with id 'item_id' of raw source is the column containing all item ids
        3. The column with id 'rating' of raw source is the column containing all the scores


        In case you would like to use the sentiment of the `review` column of the above raw source as score column,
        simply specify the appropriate `ScoreProcessor` object

        >>> ratings_raw_source = ca.CSVFile('ratings.csv')
        >>> ratings = ca.Ratings(ratings_raw_source,
        >>>                      score_column='review',
        >>>                      score_processor=ca.TextBlobSentimentAnalysis())

    Args:
        source: Source containing the raw interaction frame
        user_id_column: Name or positional index of the field of the raw source representing *users* column
        item_id_column: Name or positional index of the field of the raw source representing *items* column
        score_column: Name or positional index of the field of the raw source representing *score* column
        timestamp_column: Name or positional index of the field of the raw source representing *timesamp* column
        score_processor: `ScoreProcessor` object which will process the `score_column` accordingly. Useful if you want
            to perform sentiment analysis on a textual column or you want to normalize all scores in $[0, 1]$ range
    """

    def __init__(self, source: RawInformationSource,
                 user_id_column: Union[str, int] = 0,
                 item_id_column: Union[str, int] = 1,
                 score_column: Union[str, int] = 2,
                 timestamp_column: Union[str, int] = None,
                 score_processor: ScoreProcessor = None):

        self._ratings_dict = self._import_ratings(source, user_id_column, item_id_column, score_column,
                                                  timestamp_column, score_processor)

    @property
    @functools.lru_cache(maxsize=128)
    def user_id_column(self) -> list:
        """
        Getter for the user id column. This will return the user column "as is", so it will contain duplicate users.
        Use set(user_id_column) to get all
        unique users

        Returns:
            Users column with duplicates
        """
        return [interaction.user_id for interaction in self]

    @property
    @functools.lru_cache(maxsize=128)
    def item_id_column(self) -> list:
        """
        Getter for the user id column. This will return the item column "as is", so it will contain duplicate items.
        Use set(item_id_column) to get all unique users

        Returns:
            Items column with duplicates
        """
        return [interaction.item_id for interaction in self]

    @property
    @functools.lru_cache(maxsize=128)
    def score_column(self) -> list:
        """
        Getter for the score column. This will return the score column "as is".

        Returns:
            Score column
        """
        return [interaction.score for interaction in self]

    @property
    @functools.lru_cache(maxsize=128)
    def timestamp_column(self) -> list:
        """
        Getter for the timestamp column. This will return the score column "as is". If no timestamp is present then an
        empty list is returned

        Returns:
            Timestamp column or empty list if no timestamp is present
        """
        return [interaction.timestamp for interaction in self if interaction.timestamp is not None]

    @Handler_ScoreNotFloat
    def _import_ratings(self, source: RawInformationSource,
                        user_column: Union[str, int],
                        item_column: Union[str, int],
                        score_column: Union[str, int],
                        timestamp_column: Union[str, int],
                        score_processor: ScoreProcessor):
        ratings_dict = defaultdict(list)

        with get_progbar(source) as pbar:

            pbar.set_description(desc="Importing ratings")
            for row in pbar:

                user_id = self._get_field_data(user_column, row)
                item_id = self._get_field_data(item_column, row)
                score = self._get_field_data(score_column, row)
                timestamp = None

                if score_processor is not None:
                    score = score_processor.fit(score)
                else:
                    score = float(score)

                if timestamp_column is not None:
                    timestamp = self._get_field_data(timestamp_column, row)

                ratings_dict[user_id].append(Interaction(user_id, item_id, score, timestamp))

        # re-hashing
        return dict(ratings_dict)

    def get_user_interactions(self, user_id: str, head: int = None) -> List[Interaction]:
        """
        Method which returns a list of `Interaction` objects for a single user, one for each interaction of the user.
        Then you can easily iterate and extract useful information using list comprehension

        Examples:

            So if the rating frame is the following:

            ```
            +---------+---------+-------+
            | user_id | item_id | score |
            +---------+---------+-------+
            | u1      | i1      |     4 |
            | u1      | i2      |     3 |
            | u2      | i5      |     1 |
            +---------+---------+-------+
            ```

            >>> rating_frame.get_user_interactions('u1')
            [Interaction(user_id='u1', item_id='i1', score=4),
            Interaction(user_id='u1', item_id='i2', score=3)]

            So you could easily extract all the ratings that a user has given for example:

            >>> [interaction.score for interaction in rating_frame.get_user_interactions('u1')]
            [4, 3]

            If you only want the first $k$ interactions of the user, set `head=k`. The interactions returned are the
            first $k$ according to their order of appearance in the rating frame:

            >>> rating_frame.get_user_interactions('u1', head=1)
            [Interaction(user_id='u1', item_id='i1', score=4)]

        Args:
            user_id: User for which interactions must be extracted
            head: Integer which will cut the list of interactions of the user returned. The interactions returned are
                the first $k$ according to their order of appearance

        Returns:
            List of Interaction objects of a single user

        """
        return self._ratings_dict[user_id][:head]

    def filter_ratings(self, user_list: Iterable[str]) -> Ratings:
        """
        Method which will filter the rating frame by keeping only interactions of users appearing in the `user_list`.
        This method will return a new `Ratings` object without changing the original

        Examples:

            ```title="Starting Rating object"
            +---------+---------+-------+
            | user_id | item_id | score |
            +---------+---------+-------+
            | u1      | i1      |     4 |
            | u1      | i2      |     3 |
            | u2      | i5      |     1 |
            +---------+---------+-------+
            ```

            >>> rating_frame.filter_ratings(['u1'])

            ```title="Returned Rating object"
            +---------+---------+-------+
            | user_id | item_id | score |
            +---------+---------+-------+
            | u1      | i1      |     4 |
            | u1      | i2      |     3 |
            +---------+---------+-------+
            ```

        Args:
            user_list: List of user ids that will be present in the filtered `Ratings` object

        Returns
            The filtered Ratings object which contains only interactions of selected users
        """
        filtered_ratings_generator = ((user, self._ratings_dict[user]) for user in user_list)

        return self.from_dict(filtered_ratings_generator)

    def take_head_all(self, head: int) -> Ratings:
        """
        Method which will retain only $k$ interactions for each user. The $k$ interactions retained are the first which
        appears in the rating frame.

        This method will return a new `Ratings` object without changing the original

        Examples:

            ```title="Starting Rating object"
            +---------+---------+-------+
            | user_id | item_id | score |
            +---------+---------+-------+
            | u1      | i1      |     4 |
            | u1      | i2      |     3 |
            | u2      | i5      |     1 |
            | u2      | i6      |     2 |
            +---------+---------+-------+
            ```

            >>> rating_frame.take_head_all(head=1)

            ```title="Returned Rating object"
            +---------+---------+-------+
            | user_id | item_id | score |
            +---------+---------+-------+
            | u1      | i1      |     4 |
            | u2      | i5      |     1 |
            +---------+---------+-------+
            ```

        Args:
            head: The number of interactions to retain for each user

        Returns:
            The filtered Ratings object which contains only first $k$ interactions for each user
        """

        ratings_cut_generator = ((user_id, user_ratings[:head])
                                 for user_id, user_ratings in
                                 zip(self._ratings_dict.keys(), self._ratings_dict.values()))

        return self.from_dict(ratings_cut_generator)

    # @Handler_ScoreNotFloat
    # def add_score_column(self, score_column: Union[str, int], column_name: str,
    #                      score_processor: ScoreProcessor = None):
    #
    #     col_to_add = [self._get_field_data(score_column, row) for row in self._source]
    #
    #     if score_processor:
    #         col_to_add = score_processor.fit(col_to_add)
    #     else:
    #         col_to_add = [float(score) for score in col_to_add]
    #
    #     self._ratings_frame[column_name] = col_to_add
    #
    #     start_ratings_user = 0
    #     for user_id, user_ratings in zip(self._ratings_dict.keys(), self._ratings_dict.values()):
    #         first_range_val = start_ratings_user
    #         second_range_val = start_ratings_user + len(user_ratings)
    #
    #         score_to_add = col_to_add[first_range_val:second_range_val]
    #         new_ratings = [rating_tuple + (added_score,) for rating_tuple, added_score in
    #                        zip(user_ratings, score_to_add)]
    #         self._ratings_dict[user_id] = new_ratings
    #
    #         start_ratings_user += len(user_ratings)

    def to_dataframe(self) -> pd.DataFrame:
        """
        Method which will convert the `Rating` object to a `pandas DataFrame object`.

        The returned DataFrame object will contain the 'user_id', 'item_id' and 'score' column and optionally the
        'timestamp' column, if at least one interaction has a timestamp.

        Returns:
            The rating frame converted to a pandas DataFrame with 'user_id', 'item_id', 'score' column and optionally
            the 'timestamp' column

        """
        will_be_frame = {'user_id': self.user_id_column,
                         'item_id': self.item_id_column,
                         'score': self.score_column}

        if len(self.timestamp_column) != 0:
            will_be_frame['timestamp'] = self.timestamp_column

        return pd.DataFrame(will_be_frame)

    def to_csv(self, output_directory: str = '.', file_name: str = 'ratings_frame', overwrite: bool = False):
        """
        Method which will save the `Ratings` object to a `csv` file

        Args:
            output_directory: directory which will contain the csv file
            file_name: Name of the csv_file
            overwrite: If set to True and a csv file exists in the same output directory with the same file name, it
                will be overwritten
        """
        Path(output_directory).mkdir(parents=True, exist_ok=True)

        file_name = get_valid_filename(output_directory, file_name, 'csv', overwrite)

        frame = self.to_dataframe()
        frame.to_csv(os.path.join(output_directory, file_name), index=False, header=True)

    @staticmethod
    def _get_field_data(field_name: Union[str, int], row: Dict):
        try:
            if isinstance(field_name, str):
                data = row[field_name]
            else:
                row_keys = list(row.keys())
                key = row_keys[field_name]
                data = row[key]

        except KeyError:
            raise KeyError("Column {} not found in the raw source".format(field_name))
        except IndexError:
            raise IndexError("Column index {} not present in the raw source".format(field_name))

        return str(data)

    @classmethod
    def from_dataframe(cls, interaction_frame: pd.DataFrame,
                       user_column: Union[str, int] = 0,
                       item_column: Union[str, int] = 1,
                       score_column: Union[str, int] = 2,
                       timestamp_column: Union[str, int] = None) -> Ratings:
        """
        Class method which allows to instantiate a `Ratings` object by using an existing pandas DataFrame

        **If** the pandas DataFrame contains users, items and ratings in this order,
        no additional parameters are needed, **otherwise** the mapping must be explicitly specified using:

        * **'user_id'** column,
        * **'item_id'** column,
        * **'score'** column

        Check documentation of the `Ratings` class for examples on mapping columns explicitly, the functioning is the
        same

        Examples:

            >>> ratings_df = pd.DataFrame({'user_id': ['u1', 'u1', 'u1'],
            >>>                            'item_id': ['i1', 'i2', 'i3'],
            >>>                            'score': [4, 3, 3])
            >>> Ratings.from_dataframe(ratings_df)

        Args:
            interaction_frame: pandas DataFrame which represents the original interactions frame
            user_column: Name or positional index of the field of the DataFrame representing *users* column
            item_column: Name or positional index of the field of the DataFrame representing *items* column
            score_column: Name or positional index of the field of the DataFrame representing *score* column
            timestamp_column: Name or positional index of the field of the raw source representing *timesamp* column

        Returns:
            `Ratings` object instantiated thanks to an existing Pandas DataFrame
        """

        def get_value_row_df(row, column, dtype):
            try:
                if isinstance(column, str):
                    value = row[column]
                else:
                    # it's an int, so we get the column id and then we get the corresponding value in the row
                    key_dict = interaction_frame.columns[column]
                    value = row[key_dict]
            except (KeyError, IndexError) as e:
                if isinstance(e, KeyError):
                    raise KeyError(f"Column {column} not found in interaction frame!")
                else:
                    raise IndexError(f"Column {column} not found in interaction frame!")

            return dtype(value) if value is not None else None

        obj = cls.__new__(cls)  # Does not call __init__
        super(Ratings, obj).__init__()  # Don't forget to call any polymorphic base class initializers

        ratings_dict = defaultdict(list)

        if not interaction_frame.empty:
            for row in interaction_frame.to_dict(orient='records'):
                user_id = get_value_row_df(row, user_column, str)
                item_id = get_value_row_df(row, item_column, str)
                score = get_value_row_df(row, score_column, float)
                timestamp = get_value_row_df(row, timestamp_column, str) if timestamp_column is not None else None

                ratings_dict[user_id].append(Interaction(user_id, item_id, score, timestamp))

        obj._ratings_dict = dict(ratings_dict)
        return obj

    @classmethod
    def from_list(cls, interaction_list: Union[List[Interaction], Iterator]) -> Ratings:
        """
        Class method which allows to instantiate a `Ratings` object by using an existing list containing `Interaction`
        objects or its generator

        Examples:

            >>> interactions_list = [Interaction('u1', 'i1', 5), Interaction('u2', 'i1', 4)]
            >>> Ratings.from_list(interactions_list)

        Args:
            interaction_list: List containing `Interaction` objects or its generator

        Returns:
            `Ratings` object instantiated thanks to an existing interaction list
        """
        obj = cls.__new__(cls)  # Does not call __init__
        super(Ratings, obj).__init__()  # Don't forget to call any polymorphic base class initializers

        ratings_dict = defaultdict(list)
        for interaction in interaction_list:
            ratings_dict[interaction.user_id].append(interaction)

        obj._ratings_dict = dict(ratings_dict)
        return obj

    @classmethod
    def from_dict(cls, interaction_dict: Union[Dict[str, List[Interaction]], Iterator]) -> Ratings:
        """
        Class method which allows to instantiate a `Ratings` object by using an existing dictionary containing
        user_id as keys and lists of `Interaction` objects as value

        Examples:

            >>> interactions_dict = {'u1': [Interaction('u1', 'i2', 4), Interaction('u1', 'i3', 3)],
            >>>                      'u2': [Interaction('u2', 'i2', 5)]}
            >>> Ratings.from_dict(interactions_dict)

        Args:
            interaction_dict: Dictionary containing user_id as keys and lists of `Interaction` objets as values
                or its generator

        Returns:
            `Ratings` object instantiated thanks to an existing dictionary
        """
        obj = cls.__new__(cls)  # Does not call __init__
        super(Ratings, obj).__init__()  # Don't forget to call any polymorphic base class initializers

        obj._ratings_dict = dict(interaction_dict)
        return obj

    def __len__(self):
        # all columns have same length, so only one is needed in order
        # to check what is the length
        return len(self.user_id_column)

    def __str__(self):
        return str(self.to_dataframe())

    def __repr__(self):
        return repr(self._ratings_dict)

    def __iter__(self):
        """
        The `Ratings` object can be iterated over and each iteration will return an `Interaction` object

        Examples:

            ```title="Rating object to iterate"
            +---------+---------+-------+
            | user_id | item_id | score |
            +---------+---------+-------+
            | u1      | i1      |     4 |
            | u2      | i5      |     1 |
            +---------+---------+-------+
            ```

            >>> # for simplicity we stop after the first iteration
            >>> for interaction in ratings:
            >>>     first_interaction = interaction
            >>>     break
            >>> first_interaction
            Interaction('u1', 'i1', 4)
        """
        yield from itertools.chain.from_iterable(self._ratings_dict.values())


class RatingsLowMemory:

    def __init__(self, source: RawInformationSource,
                 user_id_column: Union[str, int] = 0,
                 item_id_column: Union[str, int] = 1,
                 score_column: Union[str, int] = 2,
                 timestamp_column: Union[str, int] = None,
                 score_processor: ScoreProcessor = None):

        rat = pd.DataFrame(source, dtype=str)
        self._ratings_dict = self._import_ratings(rat, user_id_column, item_id_column, score_column,
                                                  timestamp_column, score_processor)

    @property
    @functools.lru_cache(maxsize=128)
    def user_id_column(self) -> list:
        return self._ratings_dict.index.get_level_values('user_id').tolist()

    @property
    @functools.lru_cache(maxsize=128)
    def item_id_column(self) -> list:
        return self._ratings_dict.index.get_level_values('item_id').tolist()

    @property
    @functools.lru_cache(maxsize=128)
    def score_column(self) -> list:
        return self._ratings_dict['score'].tolist()

    @property
    @functools.lru_cache(maxsize=128)
    def timestamp_column(self) -> list:
        timestamp_list = self._ratings_dict['timestamp'].tolist()
        return timestamp_list if all(timestamp is not None for timestamp in timestamp_list) else []

    @Handler_ScoreNotFloat
    def _import_ratings(self, rat: pd.DataFrame,
                        user_column: Union[str, int],
                        item_column: Union[str, int],
                        score_column: Union[str, int],
                        timestamp_column: Union[str, int],
                        score_processor: ScoreProcessor):
        """
        Imports the ratings from the source and stores in a dataframe

        Returns:
            ratings_frame: pd.DataFrame
        """
        if isinstance(user_column, int):
            user_column = rat.columns[user_column]
        if isinstance(item_column, int):
            item_column = rat.columns[item_column]
        if isinstance(score_column, int):
            score_column = rat.columns[score_column]
        if isinstance(timestamp_column, int):
            timestamp_column = rat.columns[timestamp_column]
        elif timestamp_column is None:
            rat['timestamp'] = None
            timestamp_column = 'timestamp'

        index = pd.MultiIndex.from_tuples(zip(rat[user_column].values, rat[item_column].values),
                                          names=["user_id", "item_id"])

        rat = rat[[score_column, timestamp_column]].set_index(index)

        rat.columns = ['score', 'timestamp']

        rat['score'] = pd.to_numeric(rat['score'])

        return rat

    def get_user_interactions(self, user_id: str, head: int = None):
        user_rat = self._ratings_dict.loc[user_id][:head]

        user_rat = [Interaction(user_id, index_item, row[0], row[1])
                    for index_item, row in zip(user_rat.index, user_rat.values)]

        return user_rat

    def filter_ratings(self, user_list: Iterable[str]):
        filtered_df = self._ratings_dict.loc[
            (self._ratings_dict.index.get_level_values('user_id').isin(set(user_list)))]

        filtered_df = filtered_df.reset_index(drop=False)

        return self.from_dataframe(filtered_df, user_column='user_id',
                                   item_column='item_id',
                                   score_column='score',
                                   timestamp_column='timestamp')

    def take_head_all(self, head: int):

        filtered_df = self._ratings_dict.groupby(level='user_id').head(head)

        filtered_df = filtered_df.reset_index(drop=False)

        return self.from_dataframe(filtered_df, user_column='user_id',
                                   item_column='item_id',
                                   score_column='score',
                                   timestamp_column='timestamp')

    @classmethod
    def from_dataframe(cls, interaction_frame: pd.DataFrame,
                       user_column: Union[str, int] = 0,
                       item_column: Union[str, int] = 1,
                       score_column: Union[str, int] = 2,
                       timestamp_column: Union[str, int] = None):

        obj = cls.__new__(cls)  # Does not call __init__
        super(RatingsLowMemory, obj).__init__()  # Don't forget to call any polymorphic base class initializers

        ratings_dict = cls._import_ratings(obj, interaction_frame, user_column, item_column, score_column,
                                           timestamp_column, None)

        obj._ratings_dict = ratings_dict
        return obj

    @classmethod
    def from_list(cls, interaction_list: List[Interaction]):

        obj = cls.__new__(cls)  # Does not call __init__
        super(RatingsLowMemory, obj).__init__()  # Don't forget to call any polymorphic base class initializers

        user_column_iterator = (interaction.user_id for interaction in interaction_list)
        item_column_iterator = (interaction.item_id for interaction in interaction_list)

        index = pd.MultiIndex.from_tuples(zip(user_column_iterator, item_column_iterator),
                                          names=["user_id", "item_id"])

        score_timestamp_iterator = ({'score': interaction.score, 'timestamp': interaction.timestamp}
                                    for interaction in interaction_list)

        rat = pd.DataFrame(score_timestamp_iterator, index=index)

        obj._ratings_dict = rat
        return obj

    @classmethod
    def from_dict(cls, interaction_dict: Dict[str, List[Interaction]]):
        obj = cls.__new__(cls)  # Does not call __init__
        super(RatingsLowMemory, obj).__init__()  # Don't forget to call any polymorphic base class initializers

        obj._ratings_dict = dict(interaction_dict)
        return obj

    def __iter__(self):
        yield from itertools.chain.from_iterable(self.get_user_interactions(user_id)
                                                 for user_id in
                                                 self._ratings_dict.index.get_level_values('user_id').unique())

    def __len__(self):
        return len(self._ratings_dict)


# Aliases for the Ratings class

class Prediction(Ratings):
    """
    This class is just an alias for the `Ratings` class, it has exactly same functionalities
    """
    pass


class Rank(Ratings):
    """
    This class is just an alias for the `Ratings` class, it has exactly same functionalities
    """
    pass



import os
import pickle
import lzma
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import List, Set, Union, Iterable, Dict

from clayrs.recsys.graphs.graph_metrics import GraphMetrics
from clayrs.content_analyzer.ratings_manager.ratings import Interaction, Ratings


class Node(ABC):
    """
    Abstract class that generalizes the concept of a Node

    The Node stores the actual value of the node in the 'value' attribute

    If another type of Node must be added to the graph (EX. Context), create a subclass for
    this class and create appropriate methods for the new Node in the Graph class
    """

    def __init__(self, value: str):
        self.__value = value

    @property
    def value(self):
        return self.__value

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        # If we are comparing self to an instance of the 'Node' class
        # then compare the value stored of self with the value stored
        # of the other instance and check if they are of the same class (UserNode == UserNode),
        # else compare the stored value of self directly to the other object
        return_val = False
        if isinstance(other, Node):
            return self.value == other.value and type(self) is type(other)

        return return_val

    def __lt__(self, other):
        if isinstance(other, Node):
            return self.value < other.value

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class UserNode(Node):
    """
    Class that represents 'user' nodes

    Args:
        value (object): the value to store in the node
    """

    def __init__(self, value: str):
        super().__init__(value)

    def __str__(self):
        return "User " + str(self.value)

    def __repr__(self):
        return f"UserNode({self.value})"


class ItemNode(Node):
    """
    Class that represents 'item' nodes

    Args:
        value (object): the value to store in the node
    """

    def __init__(self, value: str):
        super().__init__(value)

    def __str__(self):
        return "Item " + str(self.value)

    def __repr__(self):
        return f"ItemNode({self.value})"


class PropertyNode(Node):
    """
    Class that represents 'property' nodes

    Args:
        value (object): the value to store in the node
    """

    def __init__(self, value: str):
        super().__init__(value)

    def __str__(self):
        return "Property " + str(self.value)

    def __repr__(self):
        return f"PropertyNode({self.value})"


class Graph(ABC):
    """
    Abstract class that generalizes the concept of a Graph

    Every Graph "is born" from a dataframe which contains
    """

    @property
    @abstractmethod
    def user_nodes(self) -> Set[UserNode]:
        """
        Returns a set of 'user' nodes
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def item_nodes(self) -> Set[ItemNode]:
        """
        Returns a set of 'item' nodes'
        """
        raise NotImplementedError

    @abstractmethod
    def add_node(self, node: Union[Node, List[Node]]):
        """
        Add a 'item' node to the graph
        """
        raise NotImplementedError

    @abstractmethod
    def add_link(self, start_node: Union[Node, List[Node]], final_node: Union[Node, List[Node]],
                 weight: float = None, label: str = 'score', timestamp: str = None):
        """
        Adds an edge between the 'start_node' and the 'final_node',
        Both nodes must be present in the graph otherwise no link is created
        """
        raise NotImplementedError

    @abstractmethod
    def remove_link(self, start_node: Node, final_node: Node):
        """
        Remove the edge between the 'start_node' and the 'final_node',
        If there's no edge between the nodes, a warning is printed
        """
        raise NotImplementedError

    @abstractmethod
    def get_link_data(self, start_node: Node, final_node: Node):
        """
        Get data of the link between two nodes
        It can be None if does not exist
        """
        raise NotImplementedError

    @abstractmethod
    def node_exists(self, node: Node) -> bool:
        """
        Returns True if node exists in the graph, false otherwise
        """
        raise NotImplementedError

    @abstractmethod
    def remove_node(self, node: Union[Node, Iterable[Node]]):
        """
        PRIVATE USAGE ONLY

        Used in the Feature Selection process to remove certain nodes from the graph

        Args:
            nodes_to_remove (Iterable): iterable object containing the nodes to remove from the graph
        """
        raise NotImplementedError

    def copy(self):
        """
        Make a deep copy the graph
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo = {id(self): result}
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result

    def serialize(self, output_directory: str = ".", file_name: str = 'graph.xz'):
        """
        Serialize the graph with the pickle.dump() method

        Args:
            output_directory (str): location where the graph will be serialized
            file_name (str): name of the file which will contain the graph serialized
        """
        if not os.path.exists(output_directory):
            os.mkdir(output_directory)

        if not file_name.endswith('.xz'):
            file_name += '.xz'

        path = os.path.join(output_directory, file_name)
        with lzma.open(path, 'wb') as f:
            pickle.dump(self, f)


class BipartiteDiGraph(Graph, GraphMetrics):
    """
    Abstract class that generalizes the concept of a BipartiteGraph

    A BipartiteGraph is a Graph containing only 'user' and 'item' nodes.

    Attributes:
        source_frame (pandas.DataFrame): must contains at least 'from_id', 'to_id', 'score' columns. The graph will be
            generated from this DataFrame
        default_score_label (str): the default label of the link between two nodes.
            Default is 'score'
        default_weight (float): the default value with which a link will be weighted
            Default is 0.5
    """

    @abstractmethod
    def get_predecessors(self, node: Node) -> List[Node]:
        """
        Get all predecessors of a node
        """
        raise NotImplementedError

    @abstractmethod
    def get_successors(self, node: Node) -> List[Node]:
        """
        Get all successors of a node
        """
        raise NotImplementedError

    # will only contain users and items
    def to_ratings(self):

        node_list = list(self.user_nodes)

        interaction_list = [Interaction(node.value, succ.value,
                                        float(self.get_link_data(node, succ).get('weight')))
                            for node in node_list
                            for succ in self.get_successors(node)
                            if isinstance(succ, ItemNode) and self.get_link_data(node, succ).get('weight')]

        return Ratings.from_list(interaction_list)


class TripartiteDiGraph(BipartiteDiGraph):
    """
    Abstract class that generalize the concept of a TripartiteGraph

    A TripartiteGraph is a Graph containing 'user', 'item' and 'property' nodes, but the latter ones
    are only allowed to be linked to 'item' nodes.

    Attributes:
        source_frame (pandas.DataFrame): must contains at least 'from_id', 'to_id', 'score' columns. The graph will be
            generated from this DataFrame
        item_contents_dir (str): the path containing items serialized
        item_exo_representation (str): the exogenous representation we want to extract properties from
        item_exo_properties (list): the properties we want to extract from the exogenous representation
        default_score_label (str): the default label of the link between two nodes.
            Default is 'score'
        default_weight (float): the default value with which a link will be weighted
            Default is 0.5
    """

    @property
    @abstractmethod
    def property_nodes(self) -> Set[Node]:
        """
        Returns a set of 'property' nodes
        """
        raise NotImplementedError

    @abstractmethod
    def add_node(self, node: Union[Node, List[Node]]):
        raise NotImplementedError

    @abstractmethod
    def add_node_with_prop(self, node: Union[ItemNode, List[ItemNode]], item_exo_properties: Union[Dict, set],
                           item_contents_dir: str,
                           item_id: Union[str, List[str]] = None):
        raise NotImplementedError


class FullDiGraph(TripartiteDiGraph):
    """
    Abstract class that generalize the concept of a FullGraph

    A FullGraph is a Graph containing 'user', 'item', 'property' nodes, and every other type of node that may be
    implemented, with no restrictions in linking.

    Attributes:
        source_frame (pandas.DataFrame): must contains at least 'from_id', 'to_id', 'score' columns. The graph will be
            generated from this DataFrame
        user_contents_dir (str): the path containing users serialized
        item_contents_dir (str): the path containing items serialized
        user_exo_representation (str): the exogenous representation we want to extract properties from for the users
        user_exo_properties (list): the properties we want to extract from the exogenous representation for the users
        item_exo_representation (str): the exogenous representation we want to extract properties from for the items
        item_exo_properties (list): the properties we want to extract from the exogenous representation for the items
        default_score_label (str): the default label of the link between two nodes.
            Default is 'score'
        default_weight (float): the default value with which a link will be weighted
            Default is 0.5
    """

    @abstractmethod
    def add_node_with_prop(self, node: Union[Node, List[Node]], exo_properties: Union[Dict, set],
                           contents_dir: str,
                           content_filename: Union[str, List[str]] = None):
        raise NotImplementedError



from __future__ import annotations
import numpy as np
from sklearn.feature_extraction import DictVectorizer
from sklearn.feature_extraction.text import TfidfVectorizer
from typing import List, Union, Mapping, Iterable, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.content_analyzer.information_processor.information_processor import InformationProcessor
    from clayrs.content_analyzer.raw_information_source import RawInformationSource

from clayrs.content_analyzer.field_content_production_techniques.field_content_production_technique \
    import TfIdfTechnique
from clayrs.content_analyzer.memory_interfaces.text_interface import KeywordIndex
from clayrs.content_analyzer.utils.check_tokenization import check_tokenized, check_not_tokenized
from clayrs.utils.const import logger


class SkLearnTfIdf(TfIdfTechnique):
    """
    Class that produces a sparse vector for each content representing the tf-idf scores of its terms using SkLearn.

    Please refer to [its documentation](https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html)
    for more information about how it's computed

    Args:
        max_df:
            When building the vocabulary ignore terms that have a document
            frequency strictly higher than the given threshold (corpus-specific
            stop words).
            If float in range [0.0, 1.0], the parameter represents a proportion of
            documents, integer absolute counts.
            This parameter is ignored if vocabulary is not None.

        min_df:
            When building the vocabulary ignore terms that have a document
            frequency strictly lower than the given threshold. This value is also
            called cut-off in the literature.
            If float in range of [0.0, 1.0], the parameter represents a proportion
            of documents, integer absolute counts.
            This parameter is ignored if vocabulary is not None.

        max_features:
            If not None, build a vocabulary that only consider the top
            max_features ordered by term frequency across the corpus.

            This parameter is ignored if vocabulary is not None.

        vocabulary:
            Either a Mapping (e.g., a dict) where keys are terms and values are
            indices in the feature matrix, or an iterable over terms. If not
            given, a vocabulary is determined from the input documents.

        binary:
            If True, all non-zero term counts are set to 1. This does not mean
            outputs will have only 0/1 values, only that the tf term in tf-idf
            is binary. (Set idf and normalization to False to get 0/1 outputs).

        dtype:
            Precision of the tf-idf scores

        norm:
            Each output row will have unit norm, either:

            - 'l2': Sum of squares of vector elements is 1. The cosine
              similarity between two vectors is their dot product when l2 norm has
              been applied.
            - 'l1': Sum of absolute values of vector elements is 1.
              See :func:`preprocessing.normalize`.

        use_idf:
            Enable inverse-document-frequency reweighting. If False, idf(t) = 1.

        smooth_idf:
            Smooth idf weights by adding one to document frequencies, as if an
            extra document was seen containing every term in the collection
            exactly once. Prevents zero divisions.

        sublinear_tf:
            Apply sublinear tf scaling, i.e. replace tf with 1 + log(tf).
    """
    def __init__(self, max_df: Union[float, int] = 1.0, min_df: Union[float, int] = 1, max_features: int = None,
                 vocabulary: Union[Mapping, Iterable] = None, binary: bool = False, dtype: Callable = np.float64,
                 norm: str = 'l2', use_idf: bool = True, smooth_idf: bool = True, sublinear_tf: bool = False):

        super().__init__()
        self._sk_vectorizer = TfidfVectorizer(max_df=max_df, min_df=min_df, max_features=max_features,
                                              vocabulary=vocabulary, binary=binary, dtype=dtype,
                                              norm=norm, use_idf=use_idf, smooth_idf=smooth_idf,
                                              sublinear_tf=sublinear_tf)

    def dataset_refactor(self, information_source: RawInformationSource, field_name: str,
                         preprocessor_list: List[InformationProcessor]) -> int:
        # Creates a corpus structure, a list of string where each string is a document.
        # Then calls TfIdfVectorizer on this collection, obtaining term-document tf-idf matrix,
        # the corpus is then deleted

        corpus = []
        logger.info(f"Computing tf-idf with {str(self)}")
        for raw_content in information_source:
            processed_field_data = self.process_data(raw_content[field_name], preprocessor_list)

            processed_field_data = check_not_tokenized(processed_field_data)
            corpus.append(processed_field_data)

        self._tfidf_matrix = self._sk_vectorizer.fit_transform(corpus)
        self._feature_names = self._sk_vectorizer.get_feature_names_out()

        return self._tfidf_matrix.shape[0]

    def __str__(self):
        return "SkLearnTfIdf"

    def __repr__(self):
        return f"SkLearnTfIdf(max_df={self._sk_vectorizer.max_df}, min_df={self._sk_vectorizer.min_df}, " \
               f"max_features={self._sk_vectorizer.max_features}, vocabulary={self._sk_vectorizer.vocabulary}, " \
               f"binary={self._sk_vectorizer.binary}, dtype={self._sk_vectorizer.dtype}, " \
               f"norm={self._sk_vectorizer.norm}, use_idf={self._sk_vectorizer.use_idf}, " \
               f"smooth_idf={self._sk_vectorizer.smooth_idf}, sublinear_tf={self._sk_vectorizer.sublinear_tf})"


class WhooshTfIdf(TfIdfTechnique):
    """
    Class that produces a sparse vector for each content representing the tf-idf scores of its terms using Whoosh

    The tf-idf computation formula is:

    $$
    tf \mbox{-} idf = (1 + log10(tf)) * log10(idf)
    $$

    """

    def __init__(self):
        super().__init__()

    def dataset_refactor(self, information_source: RawInformationSource, field_name: str,
                         preprocessor_list: List[InformationProcessor]):
        # Saves the processed data in a index that will be used for frequency calculation

        logger.info(f"Computing tf-idf with {str(self)}")
        index = KeywordIndex(f'./tf_idf_{field_name}')
        index.init_writing(True)
        dataset_len = 0
        for raw_content in information_source:
            index.new_content()
            processed_field_data = self.process_data(raw_content[field_name], preprocessor_list)

            processed_field_data = check_tokenized(processed_field_data)
            index.new_field(field_name, processed_field_data)
            index.serialize_content()
            dataset_len += 1

        index.stop_writing()

        tfidf_dicts = [index.get_tf_idf(field_name, i) for i in range(dataset_len)]
        index.delete()

        vectorizer = DictVectorizer(sparse=True)
        self._tfidf_matrix = vectorizer.fit_transform(tfidf_dicts)
        self._feature_names = vectorizer.get_feature_names_out()

        return dataset_len

    def __str__(self):
        return "WhooshTfIdf"

    def __repr__(self):
        return "WhooshTfIdf()"



from unittest import TestCase

from clayrs.content_analyzer.ratings_manager.sentiment_analysis import TextBlobSentimentAnalysis
from textblob import TextBlob
import numpy as np


class TestTextBlobSentimentAnalysis(TestCase):
    def test_fit(self):
        text_reviews = ['good item', 'it was awful', 'pretty good', 'extraordinary', 'too much expensive']

        result = [TextBlobSentimentAnalysis().fit(text) for text in text_reviews]
        expected = [TextBlob(field_data).sentiment.polarity for field_data in text_reviews]

        self.assertEqual(expected, result)

        result_rounded = [TextBlobSentimentAnalysis(decimal_rounding=4).fit(text) for text in text_reviews]
        expected_rounded = [np.round(TextBlob(field_data).sentiment.polarity, 4) for field_data in text_reviews]

        self.assertEqual(expected_rounded, result_rounded)



import itertools
import string
from typing import List
import re

import nltk

from nltk import sent_tokenize
from nltk.tokenize.toktok import ToktokTokenizer
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
from nltk.corpus import wordnet
from nltk.stem.snowball import SnowballStemmer

from clayrs.content_analyzer.information_processor.information_processor import NLP
from clayrs.content_analyzer.utils.check_tokenization import check_not_tokenized


class NLTK(NLP):
    """
    Interface to the NLTK library for natural language processing features.
    
    Examples:

        * Strip multiple whitespaces from running text
        >>> nltk_obj = NLTK(strip_multiple_whitespaces=True)
        >>> nltk_obj.process('This   has  a lot  of   spaces')
        ['This', 'has', 'a', 'lot', 'of', 'spaces']

        * Remove punctuation from running text
        >>> nltk_obj = NLTK(remove_punctuation=True)
        >>> nltk_obj.process("Hello there. How are you? I'm fine, thanks.")
        ["Hello", "there", "How", "are", "you", "I", "m", "fine", "thanks"]

        * Remove stopwords using from running text
        >>> nltk_obj = NLTK(stopwords_removal=True)
        >>> nltk_obj.process("The striped bats are hanging on their feet for the best")
        ["striped", "bats", "hanging", "feet", "best"]

        * Replace URL with a normalized token `<URL>`
        >>> nltk_obj = NLTK(url_tagging=True)
        >>> nltk_obj.process("This is facebook http://facebook.com and github https://github.com")
        ['This', 'is', 'facebook', '<URL>', 'and', 'github', '<URL>']

        * Perform lemmatization on running text
        >>> nltk_obj = NLTK(lemmatization=True)
        >>> nltk_obj.process("The striped bats are hanging on their feet for best")
        ["The", "strip", "bat", "be", "hang", "on", "their", "foot", "for", "best"]

        * Perform stemming on running text
        >>> nltk_obj = NLTK(stemming=True)
        >>> nltk_obj.process("These unbelievable abnormous objects")
        ['these', 'unbeliev', 'abnorm', 'object']

        * Label each token in the running text with its POS tag
        >>> nltk_obj = NLTK(pos_tag=True)
        >>> nltk_obj.process("Facebook was fined by Hewlett Packard for spending 100€")
        ['Facebook_NNP', 'was_VBD', 'fined_VBN', 'by_IN', 'Hewlett_NNP', 'Packard_NNP', 'for_IN', 'spending_VBG',
        '100€_CD']
    
    Args:
        strip_multiple_whitespaces: If set to True, all multiple whitespaces will be reduced to only one white space
        remove_punctuation: If set to True, all punctuation from the running text will be removed
        stopwords_removal: If set to True, all stowpwords from the running text will be removed
        url_tagging: If set to True, all urls in the running text will be replaced with the `<URL>` token
        lemmatization: If set to True, each token in the running text will be brought to its lemma
        stemming: If set to True, each token in the running text will be brought to its stem
        pos_tag: If set to True, each token in the running text will be labeled with its POS tag in the form `token_TAG`
        lang: Language of the running text

    """
    _corpus_downloaded = False

    def __init__(self, *,
                 strip_multiple_whitespaces: bool = True,
                 remove_punctuation: bool = False,
                 stopwords_removal: bool = False,
                 url_tagging: bool = False,
                 lemmatization: bool = False,
                 stemming: bool = False,
                 pos_tag: bool = False,
                 lang: str = 'english'):

        if not NLTK._corpus_downloaded:
            self.__download_corpus()
            NLTK._corpus_downloaded = True

        self.stopwords_removal = stopwords_removal

        self.stemming = stemming
        self.stemmer = SnowballStemmer(language=lang)

        self.lemmatization = lemmatization
        self.lemmatizer = WordNetLemmatizer()

        self.strip_multiple_whitespaces = strip_multiple_whitespaces
        self.url_tagging = url_tagging
        self.remove_punctuation = remove_punctuation
        self.pos_tag = pos_tag
        self.__full_lang_code = lang

    def __download_corpus(self):
        try:
            nltk.data.find('corpora/stopwords')
        except LookupError:
            nltk.download('stopwords')
        try:
            nltk.data.find('punkt')
        except LookupError:
            nltk.download('punkt')
        try:
            nltk.data.find('averaged_perceptron_tagger')
        except LookupError:
            nltk.download('averaged_perceptron_tagger')
        try:
            nltk.data.find('wordnet')
        except LookupError:
            nltk.download('wordnet')
        try:
            nltk.data.find('maxent_ne_chunker')
        except LookupError:
            nltk.download('maxent_ne_chunker')
        try:
            nltk.data.find('words')
        except LookupError:
            nltk.download('words')
        try:
            nltk.data.find('omw-1.4')
        except LookupError:
            nltk.download('omw-1.4')

    def __tokenization_operation(self, text) -> List[str]:
        """
        Splits the text in one-word tokens

        Args:
             text (str): Text to split in tokens

        Returns:
             List<str>: a list of words
        """
        # EXTREMELY useful tokenizer, it tokenizes by mantaining urls as a single token
        # as well as optional <url>, <hashtag>, etc.
        # It works for sentences so we first sentence tokenize
        sentences = sent_tokenize(text, self.__full_lang_code)
        sentences_tokenized = ToktokTokenizer().tokenize_sents(sentences)
        return list(itertools.chain.from_iterable(sentences_tokenized))

    def __stopwords_removal_operation(self, text) -> List[str]:
        """
        Execute stopwords removal on input text

        Args:
            text (List<str>):

        Returns:
            filtered_sentence (List<str>): list of words from the text, without the stopwords
        """
        stop_words = set(stopwords.words(self.__full_lang_code))
        filtered_sentence = []
        for word_token in text:
            if word_token.lower() not in stop_words:
                filtered_sentence.append(word_token)

        return filtered_sentence

    def __stemming_operation(self, text) -> List[str]:
        """
        Execute stemming on input text

        Args:
            text (List<str>):

        Returns:
            stemmed_text (List<str>): List of the fords from the text, reduced to their stem version
        """
        stemmed_text = [self.stemmer.stem(word) for word in text]

        return stemmed_text

    def __lemmatization_operation(self, text) -> List[str]:
        """
        Execute lemmatization on input text

        Args:
            text (List<str>):

        Returns:
            lemmatized_text (List<str>): List of the fords from the text, reduced to their lemmatized version
        """

        def get_wordnet_pos(word):
            """
            Map POS tag to first character lemmatize() accepts
            """
            tag = nltk.pos_tag([word])[0][1][0].upper()
            tag_dict = {"J": wordnet.ADJ,
                        "N": wordnet.NOUN,
                        "V": wordnet.VERB,
                        "R": wordnet.ADV}

            return tag_dict.get(tag, wordnet.NOUN)

        lemmatized_text = []
        for word in text:
            lemmatized_text.append(self.lemmatizer.lemmatize(word, get_wordnet_pos(word)))
        return lemmatized_text

    @staticmethod
    def __pos_operation(text) -> List[str]:
        """
        Execute POS on input text

        Args:
            text (List<str>): Text containing the entities

        Returns:
            namedEnt (nltk.tree.Tree): A tree containing the bonds between the entities
        """
        text_tuples = nltk.pos_tag(text)
        text_tagged = [f"{tagged[0]}_{tagged[1]}" for tagged in text_tuples]

        return text_tagged

    @staticmethod
    def __strip_multiple_whitespaces_operation(text) -> str:
        """
        Remove multiple whitespaces on input text

        Args:
            text (str):

        Returns:
            str: input text, multiple whitespaces removed
        """
        return re.sub(' +', ' ', text)

    @staticmethod
    def __remove_punctuation(text) -> List[str]:
        """
        Punctuation removal in spacy
        Args:
            text (List[str]):
        Returns:
            string without punctuation
        """
        # remove all tokens that are not alphabetic
        cleaned_text = [word for word in text if word not in string.punctuation]
        return cleaned_text

    @staticmethod
    def __url_tagging_operation(text) -> List[str]:
        """
        Replaces urls with <URL> string on input text

        Args:
            text (List[str]):

        Returns:
            text (list<str>): input text, <URL> instead of full urls
        """
        tagged_token = []
        for token in text:
            if re.match('http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]| '
                        '[!*(), ]|(?:%[0-9a-fA-F][0-9a-fA-F]))+',
                        token):
                tagged_token.append("<URL>")
            else:
                tagged_token.append(token)

        return tagged_token

    def process(self, field_data: str) -> List[str]:
        field_data = check_not_tokenized(field_data)
        if self.strip_multiple_whitespaces:
            field_data = self.__strip_multiple_whitespaces_operation(field_data)
        field_data = self.__tokenization_operation(field_data)
        if self.remove_punctuation:
            field_data = self.__remove_punctuation(field_data)
        if self.stopwords_removal:
            field_data = self.__stopwords_removal_operation(field_data)
        if self.pos_tag:
            field_data = self.__pos_operation(field_data)
        if self.url_tagging:
            field_data = self.__url_tagging_operation(field_data)
        if self.lemmatization:
            field_data = self.__lemmatization_operation(field_data)
        if self.stemming:
            field_data = self.__stemming_operation(field_data)
        return field_data

    def __eq__(self, other):
        if isinstance(other, NLTK):
            return self.strip_multiple_whitespaces == other.strip_multiple_whitespaces and \
                   self.remove_punctuation == other.remove_punctuation and \
                   self.stopwords_removal == other.stopwords_removal and \
                   self.url_tagging == other.url_tagging and \
                   self.lemmatization == other.lemmatization and \
                   self.stemming == other.stemming and \
                   self.pos_tag == other.pos_tag and \
                   self.__full_lang_code == other.__full_lang_code
        return False

    def __str__(self):
        return "NLTK"

    def __repr__(self):
        return f'NLTK(strip_multiple_whitespace={self.strip_multiple_whitespaces}, ' \
               f'remove_punctuation={self.remove_punctuation}, stopwords_removal={self.stopwords_removal}, ' \
               f'url_tagging={self.url_tagging}, lemmatization={self.lemmatization}, stemming={self.stemming}, ' \
               f'pos_tag={self.pos_tag}, ' \
               f'lang={self.__full_lang_code})'



from __future__ import annotations
import re
from sklearn.feature_extraction.text import CountVectorizer
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.content_analyzer.information_processor.information_processor import InformationProcessor
    from clayrs.content_analyzer.raw_information_source import RawInformationSource

from clayrs.content_analyzer.field_content_production_techniques.field_content_production_technique import \
    SynsetDocumentFrequency
from clayrs.content_analyzer.utils.check_tokenization import check_not_tokenized
from clayrs.utils.context_managers import get_progbar


class PyWSDSynsetDocumentFrequency(SynsetDocumentFrequency):
    """
    Class that produces a sparse vector for each content representing the document frequency of each synset found inside
    the document. The synsets are computed thanks to ***PyWSD*** library.

    Consider this textual representation:
    ```
    content1: "After being trapped in a jungle board game for 26 years"
    content2: "After considering jungle County, it was trapped in a jungle"
    ```

    This technique will produce the following sparse vectors:

    ```
    # vocabulary of the features
    vocabulary = {'trap.v.04': 4, 'jungle.n.03': 2, 'board.n.09': 0,
                  'plot.n.01': 3, 'twenty-six.s.01': 5,
                  'year.n.03': 7, 'view.v.02': 6, 'county.n.02': 1}

    content1:
        (0, 4)	1
        (0, 2)	1
        (0, 0)	1
        (0, 3)	1
        (0, 5)	1
        (0, 7)	1

    content2:
        (0, 4)	1
        (0, 2)	2
        (0, 6)	1
        (0, 1)	1
    ```

    """
    def __init__(self):
        # The import is here since pywsd has a long warm up phase that should affect the computation
        # only when effectively instantiated
        from pywsd import disambiguate

        self.disambiguate = disambiguate
        super().__init__()

    def dataset_refactor(self, information_source: RawInformationSource, field_name: str,
                         preprocessor_list: List[InformationProcessor]):

        all_synsets = []
        with get_progbar(information_source) as pbar:
            pbar.set_description("Computing synset frequency with wordnet")
            for raw_content in pbar:
                processed_field_data = self.process_data(raw_content[field_name], preprocessor_list)

                processed_field_data = check_not_tokenized(processed_field_data)

                synset_list = ' '.join([synset.name()
                                        for _, synset in self.disambiguate(processed_field_data)
                                        if synset is not None])
                all_synsets.append(synset_list)

        # tokenizer based on whitespaces since one document is represented as 'mysynset.id.01 mysynset.id.02 ...'
        def split_tok(text):
            return re.split("\\s+", text)

        cv = CountVectorizer(tokenizer=split_tok)
        res = cv.fit_transform(all_synsets)

        self._synset_matrix = res
        self._synset_names = cv.get_feature_names_out()

        return self._synset_matrix.shape[0]

    def __str__(self):
        return "PyWSDSynsetDocumentFrequency"

    def __repr__(self):
        return "PyWSDSynsetDocumentFrequency()"



from .combining_technique import Centroid, Sum, SingleToken
from .embedding_technique import WordEmbeddingTechnique, SentenceEmbeddingTechnique, DocumentEmbeddingTechnique, \
    Word2SentenceEmbedding, Sentence2DocEmbedding, Word2DocEmbedding, Sentence2WordEmbedding



import os
import unittest
from unittest import TestCase

from clayrs.content_analyzer.ratings_manager.ratings import Ratings, Rank
from clayrs.evaluation.metrics.error_metrics import MAE
from clayrs.evaluation.metrics.fairness_metrics import CatalogCoverage, DeltaGap
from clayrs.evaluation.metrics.plot_metrics import LongTailDistr
from clayrs.evaluation.metrics.ranking_metrics import NDCG

from clayrs.evaluation.metrics.classification_metrics import Precision
from clayrs.evaluation.eval_model import EvalModel

import pandas as pd

from test import dir_test_files

original_ratings = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1',
                 'u2', 'u2', 'u2', 'u2',
                 'u3', 'u3', 'u3', 'u3', 'u3', 'u3',
                 'u4', 'u4', 'u4', 'u4', 'u4', 'u4', 'u4',
                 'u5', 'u5', 'u5', 'u5', 'u5',
                 'u6', 'u6'],
     'item_id': ['i1', 'i2', 'i3', 'i4', 'i5', 'i6', 'i7', 'i8',
                 'i1', 'i9', 'i10', 'i11',
                 'i1', 'i12', 'i13', 'i3', 'i10', 'i14',
                 'i3', 'i10', 'i15', 'i16', 'i9', 'i17', 'i99',
                 'i10', 'i18', 'i19', 'i20', 'i21',
                 'inew_1', 'inew_2'],
     'score': [5, 4, 4, 1, 2, 3, 3, 1,
               4, 5, 1, 1,
               3, 3, 2, 1, 1, 4,
               4, 4, 5, 5, 3, 3, 3,
               3, 3, 2, 2, 1,
               4, 3]})
original_ratings = Ratings.from_dataframe(original_ratings)

train_1 = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1',  # removed last 2
                 'u2', 'u2', 'u2',  # removed last 1
                 'u3', 'u3', 'u3', 'u3',  # removed last 2
                 'u4', 'u4', 'u4', 'u4', 'u4',  # removed last 2
                 'u5', 'u5', 'u5', 'u5',  # removed last 1
                 'u6'],  # removed last 1
     'item_id': ['i1', 'i2', 'i3', 'i4', 'i5', 'i6',
                 'i1', 'i9', 'i10',
                 'i1', 'i12', 'i13', 'i3',
                 'i3', 'i10', 'i15', 'i16', 'i9',
                 'i10', 'i18', 'i19', 'i20',
                 'inew_1'],
     'score': [5, 4, 4, 1, 2, 3,
               4, 5, 1,
               3, 3, 2, 1,
               4, 4, 5, 5, 3,
               3, 3, 2, 2,
               4]})
train_1 = Ratings.from_dataframe(train_1)

truth_1 = pd.DataFrame(
    {'user_id': ['u1', 'u1',
                 'u2',
                 'u3', 'u3',
                 'u4', 'u4', 'u4',
                 'u5',
                 'u6'],
     'item_id': ['i7', 'i8',
                 'i11',
                 'i10', 'i14',
                 'i9', 'i17', 'i99',
                 'i21',
                 'inew_2'],
     'score': [3, 1,
               1,
               1, 4,
               3, 3, 3,
               1,
               3]})
truth_1 = Ratings.from_dataframe(truth_1)

train_2 = pd.DataFrame(
    {'user_id': ['u1', 'u1',  # taken last 2
                 'u2',  # taken last 1
                 'u3', 'u3',  # taken last 2
                 'u4', 'u4', 'u4',  # taken last 2
                 'u5',  # taken last 1
                 'u6'],  # taken last 1
     'item_id': ['i7', 'i8',
                 'i11',
                 'i10', 'i14',
                 'i9', 'i17', 'i99',
                 'i21',
                 'inew_2'],
     'score': [3, 1,
               1,
               1, 4,
               3, 3, 3,
               1,
               3]})
train_2 = Ratings.from_dataframe(train_2)

truth_2 = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1',
                 'u2', 'u2', 'u2',
                 'u3', 'u3', 'u3', 'u3',
                 'u4', 'u4', 'u4', 'u4', 'u4',
                 'u5', 'u5', 'u5', 'u5',
                 'u6'],
     'item_id': ['i1', 'i2', 'i3', 'i4', 'i5', 'i6',
                 'i1', 'i9', 'i10',
                 'i1', 'i12', 'i13', 'i3',
                 'i3', 'i10', 'i15', 'i16', 'i9',
                 'i10', 'i18', 'i19', 'i20',
                 'inew_1'],
     'score': [5, 4, 4, 1, 2, 3,
               4, 5, 1,
               3, 3, 2, 1,
               4, 4, 5, 5, 3,
               3, 3, 2, 2,
               4]})
truth_2 = Ratings.from_dataframe(truth_2)

# u6 is missing, just to test case in which for some users recs can't be computed
recs_1 = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1',
                 'u2', 'u2', 'u2', 'u2', 'u2',
                 'u3', 'u3', 'u3', 'u3', 'u3',
                 'u4', 'u4', 'u4', 'u4', 'u4',
                 'u5', 'u5', 'u5', 'u5', 'u5'],
     'item_id': ['i7', 'i10', 'i11', 'i12', 'i13',
                 'i11', 'i20', 'i6', 'i3', 'i4',
                 'i4', 'i5', 'i6', 'i7', 'i10',
                 'i9', 'i2', 'i3', 'i1', 'i5',
                 'i2', 'i3', 'i4', 'i5', 'i6'],
     'score': [500, 400, 300, 200, 100,
               400, 300, 200, 100, 50,
               150, 125, 110, 100, 80,
               390, 380, 360, 320, 200,
               250, 150, 190, 100, 50]})
recs_1 = Rank.from_dataframe(recs_1)

# u6 is missing, just to test case in which for some users recs can't be computed
recs_2 = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1',
                 'u2', 'u2', 'u2', 'u2', 'u2',
                 'u3', 'u3', 'u3', 'u3', 'u3',
                 'u4', 'u4', 'u4', 'u4', 'u4',
                 'u5', 'u5', 'u5', 'u5', 'u5',
                 'u6', 'u6', 'u6', 'u6', 'u6'],
     'item_id': ['i10', 'i11', 'i7', 'i12', 'i13',
                 'i20', 'i11', 'i3', 'i6', 'i4',
                 'i5', 'i4', 'i6', 'i7', 'i10',
                 'i9', 'i3', 'i1', 'i2', 'i5',
                 'i2', 'i4', 'i3', 'i5', 'i6',
                 'i10', 'i11', 'i12', 'i13', 'i14'],
     'score': [500, 400, 300, 200, 100,
               400, 300, 200, 100, 50,
               150, 125, 110, 100, 80,
               390, 380, 360, 320, 200,
               250, 150, 190, 100, 50,
               200, 100, 190, 80, 70]})
recs_2 = Rank.from_dataframe(recs_2)

pred_list = [recs_1, recs_2]
truth_list = [truth_1, truth_2]


class TestEvalModel(TestCase):

    def setUp(self) -> None:
        catalog = set(original_ratings.item_id_column)

        self.metric_list = [
            # one classification metric
            Precision(sys_average='micro'),

            # one ranking metric
            NDCG(),

            # one error metric
            MAE(),

            # two fairness metrics
            CatalogCoverage(catalog),
            DeltaGap({'a': 0.2, 'b': 0.5, 'c': 0.3},
                     user_profiles=[train_1, train_2],
                     original_ratings=original_ratings),

            # one metric which returns a plot
            LongTailDistr()
        ]

    def test_fit(self):
        em = EvalModel(pred_list, truth_list, self.metric_list)
        sys_result, user_results = em.fit()

        self.assertIsInstance(sys_result, pd.DataFrame)
        self.assertIsInstance(user_results, pd.DataFrame)

        # we should have a result for each user of the original rating frame
        self.assertEqual(set(original_ratings.user_id_column), set(user_results.index.map(str)))

        # the user result frame must contain results for each user of the Precision, NDCG, MAE (the first 3 of the
        # metric list). The other metrics do not compute result for each metric so they will not be present as columns
        # in the user_results frame
        self.assertEqual(list(user_results.columns), ['Precision - micro', 'NDCG', 'MAE'])

        # the sys_result frame must contain result of the system for each fold (2 in this case) + the mean result
        self.assertTrue(len(sys_result) == 3)
        self.assertEqual({'sys - fold1', 'sys - fold2', 'sys - mean'}, set(sys_result.index))

        # the sys result frame must contain results for the whole sys of the Precision, NDCG, MAE,
        # Catalog Coverage and DeltaGAP
        self.assertEqual(['Precision - micro', 'NDCG', 'MAE',
                          'CatalogCoverage (PredictionCov)',
                          'DeltaGap | a',
                          'DeltaGap | b',
                          'DeltaGap | c',
                          ],
                         list(sys_result.columns))

        # check that plots are generated and remove them
        self.assertTrue(os.path.isfile('long_tail_distr_truth.png'))
        os.remove('long_tail_distr_truth.png')
        self.assertTrue(os.path.isfile('long_tail_distr_truth (1).png'))
        os.remove('long_tail_distr_truth (1).png')

    def test_fit_user_list(self):
        # we compute evaluation only for a certain users

        # we must also cut user profiles of DeltaGAP by only considering u1, u2, u3
        self.metric_list.pop(4)
        self.metric_list.append(DeltaGap({'a': 0.2, 'b': 0.5, 'c': 0.3},
                                         user_profiles=[train_1.filter_ratings(['u1', 'u2', 'u3']),
                                                        train_2.filter_ratings(['u1', 'u2', 'u3'])],
                                         original_ratings=original_ratings.filter_ratings(['u1', 'u2', 'u3']))
                                )

        em = EvalModel(pred_list, truth_list, self.metric_list)
        sys_result, user_results = em.fit(['u1', 'u2', 'u3'])

        self.assertIsInstance(sys_result, pd.DataFrame)
        self.assertIsInstance(user_results, pd.DataFrame)

        self.assertEqual({'u1', 'u2', 'u3'}, set(user_results.index))

        # the user result frame must contain results for each user of the Precision, NDCG, MAE (the first 3 of the
        # metric list). The other metrics do not compute result for each metric so they will not be present as columns
        # in the user_results frame
        self.assertEqual(list(user_results.columns), ['Precision - micro', 'NDCG', 'MAE'])

        # the sys_result frame must contain result of the system for each fold (2 in this case) + the mean result
        self.assertTrue(len(sys_result) == 3)
        self.assertEqual({'sys - fold1', 'sys - fold2', 'sys - mean'}, set(sys_result.index))

        # the sys result frame must contain results for the whole sys of the Precision, NDCG, MAE,
        # Catalog Coverage and DeltaGAP
        self.assertEqual(['Precision - micro', 'NDCG', 'MAE',
                          'CatalogCoverage (PredictionCov)',
                          'DeltaGap | a',
                          'DeltaGap | b',
                          'DeltaGap | c',
                          ],
                         list(sys_result.columns))

        # check that plots are generated and remove them
        self.assertTrue(os.path.isfile('long_tail_distr_truth.png'))
        os.remove('long_tail_distr_truth.png')
        self.assertTrue(os.path.isfile('long_tail_distr_truth (1).png'))
        os.remove('long_tail_distr_truth (1).png')

    def test_fit_error(self):
        # should raise error since pred_list and truth_list must be of equal length
        with self.assertRaises(ValueError):
            pred_list_smaller = [Rank.from_dataframe(pd.DataFrame())]
            pred_list_bigger = [Ratings.from_dataframe(pd.DataFrame()), Ratings.from_dataframe(pd.DataFrame())]

            EvalModel(pred_list_smaller, pred_list_bigger, self.metric_list)


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod


class GraphMetrics(ABC):
    """
    Interface for graph metrics
    """

    @abstractmethod
    def degree_centrality(self):
        raise NotImplementedError

    @abstractmethod
    def closeness_centrality(self):
        raise NotImplementedError

    @abstractmethod
    def dispersion(self):
        raise NotImplementedError



from __future__ import annotations
import abc
from copy import deepcopy
from itertools import chain
from typing import List, TYPE_CHECKING
import pandas as pd

from scipy import sparse
from sklearn.exceptions import NotFittedError
from sklearn.feature_extraction import DictVectorizer
from sklearn.utils.validation import check_is_fitted
import numpy as np

if TYPE_CHECKING:
    from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.combining_technique import \
        CombiningTechnique
    from clayrs.content_analyzer.ratings_manager.ratings import Interaction, Prediction, Rank
    from clayrs.content_analyzer.content_representation.content import Content
    from clayrs.recsys.content_based_algorithm.contents_loader import LoadedContentsInterface


from clayrs.recsys.algorithm import Algorithm

from clayrs.recsys.content_based_algorithm.contents_loader import LoadedContentsDict


class ContentBasedAlgorithm(Algorithm):
    """
    Abstract class for the content-based algorithms

    Every Content Based Algorithm base its prediction (let it be score prediction or ranking) on representations
    serialized by the Content Analyzer. It can be used a single representation or multiple ones for a single field or
    multiple ones

    Args:
    item_field (dict): dict where the key is the name of the field
        that contains the content to use, value is the representation(s) id(s) that will be
        used for the said item. The value of a field can be a string or a list,
        use a list if you want to use multiple representations for a particular field.
    threshold (float): Threshold for the ratings. Some algorithms use this threshold in order to
        separate positive items from the negative ones, others may use only ratings that are >= than this
        threshold. See the documentation of the algorithm used for more
    """
    __slots__ = ('item_field', 'threshold', '_transformer')

    def __init__(self, item_field: dict, threshold: float):
        self.item_field: dict = self._bracket_representation(item_field)
        self.threshold: float = threshold
        self._transformer = DictVectorizer(sparse=False, sort=False)

    @staticmethod
    def _bracket_representation(item_field: dict):
        """
        Private method that brackets every representation in case the user passed a string
        instead of a list.

        EXAMPLE::
            > item_field = {'Plot': 0, 'Genre': ['tfidf', 1]}
            > print(_bracket_representation(item_field))

            > {'Plot': [0], 'Genre': ['tfidf', 1]}

        Args:
            item_field (dict): dict that may contain values that need to be bracketed

        Returns:
            The item_field passed with all values inside a list
        """
        for field in item_field:
            if not isinstance(item_field[field], list):
                item_field[field] = [item_field[field]]

        return item_field

    @staticmethod
    def _calc_mean_user_threshold(user_ratings: List[Interaction]):
        """
        Private method which simply calculates the average rating by the user given its ratings
        """
        return np.nanmean([interaction.score for interaction in user_ratings])

    def extract_features_item(self, item: Content):
        """
        Function that extracts the feature of a loaded item using the item_field parameter passed in the
        constructor.

        It extracts only the chosen representations of the chosen fields in the item loaded
        EXAMPLE:

            with item_field = {'Plot': [0], 'Genre': ['tfidf', 1]}, the function will extracts
            only the representation with '0' as internal id for the field 'Plot' and two representations
            for the field 'Genre': one with 'tfidf' as external id and the other with 1 as internal id

        Args:
            item (Content): item loaded of which we need to extract its feature

        Returns:
            A list containing all representations extracted for the item
        """
        item_bag_list = []
        if item is not None:
            for field in self.item_field:
                field_representations = self.item_field[field]

                for representation in field_representations:
                    item_bag_list.append(
                        item.get_field_representation(field, representation).value
                    )

        return item_bag_list

    def fuse_representations(self, X: list, embedding_combiner: CombiningTechnique, as_array: bool = False):
        """
        Method which transforms the X passed vectorizing if X contains dicts and merging
        multiple representations in a single one for every item in X.
        So if X = [
                    [dict, np.array, np.array]
                        ...
                    [dict, np.array, np.array]
                ]
        where every sublist contains multiple representation for a single item,
        the function returns:
        X = [
                np.array,
                ...
                np.array
            ]
        Where every row is the fused representation for the item

        In case np.array have different row size, every array will be transformed in a one dimensional one
        using the parameter embedding combiner. Check all the available combining technique to know how rows of
        a np.array can be merged into one

        Args:
            X (list): list that contains representations of the items
            embedding_combiner (CombiningTechnique): combining technique in case there are multiple
                vectors with different row size
        Returns:
            X fused and vectorized
        """
        if any(not isinstance(rep, (dict, np.ndarray, (int, float), sparse.csc_matrix)) for rep in X[0]):
            raise ValueError("You can only use representations of type: {numeric, embedding, tfidf}")

        # We check if there are dicts as representation in the first element of X,
        # since the representations are the same for all elements in X we can check
        # for dicts only in one element
        need_vectorizer = any(isinstance(rep, dict) for rep in X[0])

        if need_vectorizer:
            # IF the transformer is not fitted then we are training the model
            try:
                check_is_fitted(self._transformer)
            except NotFittedError:
                X_dicts = [rep for item in X for rep in item if isinstance(rep, dict)]
                self._transformer.fit(X_dicts)

        # In every case, we transform the input
        def single_item_fused_gen():
            for item_repr_list in X:
                single_arr = []
                for item_repr in item_repr_list:
                    if need_vectorizer and isinstance(item_repr, dict):
                        item_repr = self._transformer.transform(item_repr)
                        single_arr.append(item_repr.flatten())
                    elif isinstance(item_repr, np.ndarray):
                        if item_repr.ndim > 1:
                            item_repr = embedding_combiner.combine(item_repr)

                        single_arr.append(item_repr.flatten())
                    else:
                        # it's a float
                        single_arr.append(item_repr)

                yield single_arr

        # if a representation used is a sparse matrix, then we use scipy library to concatenate
        # otherwise, if we have all dense arrays, we use numpy. To do this check we consider the representations
        # of the first item
        first_arr = next(single_item_fused_gen())
        if any(isinstance(x, sparse.csc_matrix) for x in first_arr):
            X_vectorized = (sparse.hstack(single_arr) for single_arr in single_item_fused_gen())

            X_vectorized = sparse.vstack(X_vectorized, format='csr')

            if as_array is True:
                X_vectorized = X_vectorized.toarray()
        else:
            X_vectorized = [np.hstack(single_arr) for single_arr in single_item_fused_gen()]

            X_vectorized = np.array(X_vectorized)

        return X_vectorized

    @abc.abstractmethod
    def process_rated(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsDict):
        """
        Abstract method that processes rated items for the user.

        Every content-based algorithm processes rated items differently, it may be needed to extract features
        from the rated items and label them, extract features only from the positive ones, etc.

        The rated items processed must be stored into a private attribute of the algorithm, later used
        by the fit() method.

        Args:
            user_ratings (pd.DataFrame): DataFrame containing ratings of a single user
            available_loaded_items (LoadedContentsDict): loaded contents interface point to the items directory
        """
        raise NotImplementedError

    @abc.abstractmethod
    def fit(self):
        """
        Abstract method that fits the content-based algorithm.

        Every content based algorithm has a different fit process, it may be needed to fit a classifier,
        to build the centroid of the positive items, to build a query for the index, etc.

        It must be called after the the process_rated() method since it uses private attributes calculated
        by said method to fit the algorithm.

        The fitted object will also be stored in a private attribute.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsDict,
                filter_list: List[str] = None) -> Prediction:
        """
        |  Abstract method that predicts the rating which a user would give to items
        |  If the algorithm is not a PredictionScore Algorithm, implement this method like this:

        def predict():
            raise NotPredictionAlg

        One can specify which items must be predicted with the filter_list parameter,
        in this case ONLY items in the filter_list will be predicted.
        One can also pass items already seen by the user with the filter_list parameter.
        Otherwise, ALL unrated items will be predicted.

        Args:
            user_ratings (pd.DataFrame): DataFrame containing ratings of a single user
            available_loaded_items (LoadedContentsDict): loaded contents interface point to the items directory
            filter_list (list): list of the items to predict, if None all unrated items will be score predicted
        Returns:
            pd.DataFrame: DataFrame containing one column with the items name,
                one column with the score predicted
        """
        raise NotImplementedError

    @abc.abstractmethod
    def rank(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsInterface,
             recs_number: int = None, filter_list: List[str] = None) -> Rank:
        """
        Rank the top-n recommended items for the user. If the recs_number parameter isn't specified,
        All unrated items for the user will be ranked (or only items in the filter list, if specified).

        One can specify which items must be ranked with the `filter_list` parameter,
        in this case ONLY items in the `filter_list` parameter will be ranked.
        One can also pass items already seen by the user with the filter_list parameter.
        Otherwise, **ALL** unrated items will be ranked.

        Args:
            user_ratings: List of Interaction objects for a single user
            available_loaded_items: The LoadedContents interface which contains loaded contents
            recs_number: number of the top ranked items to return, if None all ranked items will be returned
            filter_list (list): list of the items to rank, if None all unrated items for the user will be ranked

        Returns:
            List of Interactions object in a descending order w.r.t the 'score' attribute, representing the ranking for
            a single user
        """
        raise NotImplementedError

    def _load_available_contents(self, contents_path: str, items_to_load: set = None):
        return LoadedContentsDict(contents_path, items_to_load, only_representations=self.item_field)

    def __deepcopy__(self, memo):
        # Create a new instance
        cls = self.__class__
        result = cls.__new__(cls)

        # Don't copy self reference
        memo[id(self)] = result

        # Don't copy the cache - if it exists
        if hasattr(self, "_cache"):
            memo[id(self._cache)] = self._cache.__new__(dict)

        # Get all __slots__ of the derived class
        slots = chain.from_iterable(getattr(s, '__slots__', []) for s in self.__class__.__mro__)

        # Deep copy all other attributes
        for var in slots:
            setattr(result, var, deepcopy(getattr(self, var), memo))

        # Return updated instance
        return result



from abc import ABC, abstractmethod
from typing import Union
import shutil


class InformationInterface(ABC):
    """
    Abstract class that deals with the serialization
    and deserialization of a field (of a content) data
    basing on the type of element extracted.

    Args:
        directory (str): directory where to store the serialized content and where to access for deserialization
    """

    def __init__(self, directory: str):
        self.__directory: str = directory

    def delete(self):
        shutil.rmtree(self.directory, ignore_errors=True)

    @property
    def directory(self):
        return self.__directory

    @abstractmethod
    def new_content(self):
        """
        Creates a new item, that will be serialized by the apposite method.
        """
        raise NotImplementedError

    @abstractmethod
    def new_field(self, field_name: str, field_data):
        """
        Serialize the raw data of a field.

        Args:
            field_name: name of the created field
            field_data: data to serialize
        """
        raise NotImplementedError

    @abstractmethod
    def serialize_content(self):
        """
        Add to the serialized collection the current item
        """
        raise NotImplementedError

    @abstractmethod
    def init_writing(self, delete_old: bool = False):
        """
        Set the interface in writing mode,
        tf the specified directory does not exit a new one will be created.
        If the directory exists and an object associated to the IndexInterface already exists, what happens depends on
        the delete_old attribute

        Args:
            delete_old (bool): if True, the object handled by the information interface that was in the same directory
                is destroyed and replaced; if False, the object is simply opened
        """
        raise NotImplementedError

    @abstractmethod
    def stop_writing(self):
        """
        Stop writing mode
        """
        raise NotImplementedError

    @abstractmethod
    def get_field(self, field_name: str, content_id: Union[str, int]):
        """
        Allows to retrieve the content stored in a field for a content

        Args:
            field_name (str): name of the field from which the data will be retrieved
            content_id (Union[str, int]): either the position or Id of the content that contains the specified field
        """
        raise NotImplementedError

    def __hash__(self):
        return hash(str(self.__directory))

    def __eq__(self, other):
        if isinstance(other, InformationInterface):
            return self.directory == other.directory


class ImageInterface(InformationInterface):
    """
    Future feature
    Abstract class to use when the field information is in image format.
    """

    @abstractmethod
    def new_content(self):
        raise NotImplementedError

    @abstractmethod
    def new_field(self, field_name: str, field_data):
        raise NotImplementedError

    @abstractmethod
    def serialize_content(self):
        raise NotImplementedError

    @abstractmethod
    def init_writing(self, delete_old: bool = False):
        raise NotImplementedError

    @abstractmethod
    def stop_writing(self):
        raise NotImplementedError

    @abstractmethod
    def get_field(self, field_name: str, content_id: Union[str, int]):
        raise NotImplementedError


class AudioInterface(InformationInterface):
    """
    Future feature
    Abstract class to use when the field information is in audio format.
    """

    @abstractmethod
    def new_content(self):
        raise NotImplementedError

    @abstractmethod
    def new_field(self, field_name: str, field_data):
        raise NotImplementedError

    @abstractmethod
    def serialize_content(self):
        raise NotImplementedError

    @abstractmethod
    def init_writing(self, delete_old: bool = False):
        raise NotImplementedError

    @abstractmethod
    def stop_writing(self):
        raise NotImplementedError

    @abstractmethod
    def get_field(self, field_name: str, content_id: Union[str, int]):
        raise NotImplementedError


class TextInterface(InformationInterface):
    """
    Abstract class to use when the field information is textual.
    """

    @abstractmethod
    def new_content(self):
        raise NotImplementedError

    @abstractmethod
    def new_field(self, field_name: str, field_data):
        raise NotImplementedError

    @abstractmethod
    def serialize_content(self):
        raise NotImplementedError

    @abstractmethod
    def init_writing(self, delete_old: bool = False):
        raise NotImplementedError

    @abstractmethod
    def stop_writing(self):
        raise NotImplementedError

    @abstractmethod
    def get_field(self, field_name: str, content_id: Union[str, int]):
        raise NotImplementedError

    @abstractmethod
    def query(self, string_query: str, results_number: int, mask_list: list = None,
              candidate_list: list = None, classic_similarity: bool = True) -> dict:
        raise NotImplementedError

    @abstractmethod
    def get_tf_idf(self, field_name: str, content_id: Union[str, int]):
        raise NotImplementedError



from __future__ import annotations
import itertools
from collections import Counter
from pathlib import Path
from typing import Union, Dict, TYPE_CHECKING

import matplotlib as mpl
import matplotlib.figure
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import numpy as np
import pandas as pd
import os
import matplotlib.ticker as plticker

if TYPE_CHECKING:
    from clayrs.content_analyzer import Ratings
    from clayrs.recsys.partitioning import Split

from clayrs.evaluation.utils import get_item_popularity, get_most_popular_items
from clayrs.utils.save_content import get_valid_filename
from clayrs.evaluation.metrics.fairness_metrics import GroupFairnessMetric, pop_ratio_by_user
from clayrs.evaluation.metrics.metrics import Metric
from clayrs.utils.const import logger


class PlotMetric(Metric):
    """
    A plot metric is a metric which generates a plot and saves it to the directory specified

    The plot file will be saved as `out_dir/file_name.format`

    Since multiple split could be evaluated at once, the *overwrite* parameter comes into play:
    if is set to False, file with the same name will be saved as `file_name (1).format`, `file_name (2).format`, etc.
    so that for every split a plot is generated without overwriting any file previously generated

    Args:
        out_dir (str): Directory where the plot will be saved. Default is '.', meaning that the plot will be saved
            in the same directory where the python script it's being executed
        file_name (str): Name of the plot file. Every plot metric as a default file name
        format (str): Format of the plot file. Could be 'jpg', 'svg', 'png'. Default is 'png'
        overwrite (bool): parameter which specifies if the plot saved must overwrite any file that as the same name
            ('file_name.format'). Default is False
    """

    def __init__(self, out_dir: str = '.', file_name: str = None, format: str = 'png', overwrite: bool = False):
        self.__out_dir = out_dir
        self.__file_name = file_name
        self.__format = format
        self.__overwrite = overwrite

    @property
    def file_name(self):
        return self.__file_name

    @property
    def output_directory(self):
        return self.__out_dir

    @property
    def format(self):
        return self.__format

    @property
    def overwrite(self):
        return self.__overwrite

    def save_figure(self, fig, file_name: str):
        """
        Concrete method which given the figure to save and its file name, it saves the figure in the output directory
        and with the format specified in the constructor

        Args:
            fig: figure to save
            file_name (str): name of the file to save
        """
        Path(self.output_directory).mkdir(parents=True, exist_ok=True)

        file_name = get_valid_filename(self.output_directory, file_name, self.format, self.overwrite)
        fig.savefig(os.path.join(self.output_directory, file_name))
        fig.clf()
        plt.close(fig)


class LongTailDistr(PlotMetric):
    """
    This metric generates the Long Tail Distribution plot and saves it in the output directory with the file name
    specified. The plot can be generated both for the *truth set* or the *predictions set* (based on
    the *on* parameter):

    - **on = 'truth'**: in this case the long tail distribution is useful to see which are the most popular items (the
      most rated ones)

    - **on = 'pred'**: in this case the long tail distribution is useful to see which are the most recommended items

    The plot file will be saved as `out_dir/file_name.format`

    Since multiple split could be evaluated at once, the *overwrite* parameter comes into play:
    if is set to False, file with the same name will be saved as `file_name (1).format`, `file_name (2).format`, etc.
    so that for every split a plot is generated without overwriting any file previously generated

    Args:
        out_dir: Directory where the plot will be saved. Default is '.', meaning that the plot will be saved
            in the same directory where the python script it's being executed
        file_name: Name of the plot file. Default is 'long_tail_distr'
        on: Set on which the Long Tail Distribution plot will be generated. Values accepted are 'truth' or 'pred'
        format: Format of the plot file. Could be 'jpg', 'svg', 'png'. Default is 'png'
        overwrite: parameter which specifies if the plot saved must overwrite any file that as the same name
            ('file_name.format'). Default is False

    Raises:
        ValueError: exception raised when a invalid value for the 'on' parameter is specified
    """

    def __init__(self, out_dir: str = '.', file_name: str = 'long_tail_distr', on: str = 'truth', format: str = 'png',
                 overwrite: bool = False):
        valid = {'truth', 'pred'}
        self.__on = on.lower()

        if self.__on not in valid:
            raise ValueError("on={} is not supported! Long Tail can be calculated only on:\n"
                             "{}".format(on, valid))
        super().__init__(out_dir, file_name, format, overwrite)

    def __str__(self):
        return "LongTailDistr"

    def __repr__(self):
        return f'LongTailDistr(out_dir={self.output_directory}, ' \
               f'file_name={self.file_name}, ' \
               f'on={self.__on}, ' \
               f'format={self.format}, ' \
               f'overwrite={self.overwrite})'

    def perform(self, split: Split) -> pd.DataFrame:
        if self.__on == 'truth':
            frame = split.truth
        else:
            frame = split.pred

        counts_by_item = Counter(frame.item_id_column)
        ordered_item_count_pairs = counts_by_item.most_common()

        ordered_counts = []
        labels = []
        for item_count_pair in ordered_item_count_pairs:
            labels.append(item_count_pair[0])
            ordered_counts.append(item_count_pair[1])

        x = [i for i in range(len(labels))]
        y = ordered_counts

        fig = plt.figure()
        ax = fig.add_subplot()

        if self.__on == 'truth':
            ax.set(xlabel='Item ID', ylabel='# of item ratings',
                   title='Long Tail Distribution - {}'.format(self.__on))
        else:
            ax.set(xlabel='Item ID', ylabel='# of item recommendations',
                   title='Long Tail Distribution - {}'.format(self.__on))

        ax.fill_between(x, y, color="orange", alpha=0.2)
        ax.set_xticks([])
        ax.plot(x, y)

        file_name = self.file_name + '_{}'.format(self.__on)

        self.save_figure(fig, file_name=file_name)

        return pd.DataFrame(columns=['user_id', 'item_id', 'score'])


class PopRatioProfileVsRecs(GroupFairnessMetric, PlotMetric):
    r"""
    This metric generates a plot where users are split into groups and, for every group, a boxplot comparing
    ***profile popularity ratio*** and ***recommendations popularity ratio*** is drawn

    Users are split into groups based on the *user_groups* parameter, which contains names of the groups as keys,
    and percentage of how many user must contain a group as values. For example:

        user_groups = {'popular_users': 0.3, 'medium_popular_users': 0.2, 'low_popular_users': 0.5}

    Every user will be inserted in a group based on how many popular items the user has rated (in relation to the
    percentage of users we specified as value in the dictionary):

    * users with many popular items will be inserted into the first group
    * users with niche items rated will be inserted into one of the last groups.

    In general users are grouped by $Popularity\_ratio$ in a descending order. $Popularity\_ratio$ for a single user $u$
    is defined as:

    $$
    Popularity\_ratio_u = n\_most\_popular\_items\_rated_u / n\_items\_rated_u
    $$

    The *most popular items* are the first `pop_percentage`% items of all items ordered in a descending order by
    popularity.

    The popularity of an item is defined as the number of times it is rated in the `original_ratings` parameter
    divided by the total number of users in the `original_ratings`.

    It can happen that for a particular user of a group no recommendation are available: in that case it will be skipped
    and it won't be considered in the $Popularity\_ratio$ computation of its group. In case no user of a group has recs
    available, a warning will be printed and the whole group won't be considered.

    The plot file will be saved as `out_dir/file_name.format`

    Since multiple split could be evaluated at once, the `overwrite` parameter comes into play:
    if is set to False, file with the same name will be saved as `file_name (1).format`, `file_name (2).format`, etc.
    so that for every split a plot is generated without overwriting any file previously generated

    Thanks to the 'store_frame' parameter it's also possible to store a csv containing the calculations done in order
    to build every boxplot. Will be saved in the same directory and with the same file name as the plot itself (but
    with the .csv format):

    The csv will be saved as `out_dir/file_name.csv`

    ***Please note***: once computed, the DeltaGAP class needs to be re-instantiated in case you want to compute it
    again!

    Args:
        user_groups (Dict<str, float>): Dict containing group names as keys and percentage of users as value, used to
            split users in groups. Users with more popular items rated are grouped into the first group, users with
            slightly less popular items rated are grouped into the second one, etc.
        user_profiles: one or more `Ratings` objects containing interactions of the profile of each user
            (e.g. the **train set**). It should be one for each split to evaluate!
        original_ratings: `Ratings` object containing original interactions of the dataset that will be used to
            compute the popularity of each item (i.e. the number of times it is rated divided by the total number of
            users)
        out_dir (str): Directory where the plot will be saved. Default is '.', meaning that the plot will be saved
            in the same directory where the python script it's being executed
        file_name (str): Name of the plot file. Default is 'pop_ratio_profile_vs_recs'
        pop_percentage (float): How many (in percentage) 'most popular items' must be considered. Default is 0.2
        store_frame (bool): True if you want to store calculations done in order to build every boxplot in a csv file,
            False otherwise. Default is set to False
        format (str): Format of the plot file. Could be 'jpg', 'svg', 'png'. Default is 'png'
        overwrite (bool): parameter which specifies if the plot saved must overwrite any file that as the same name
            ('file_name.format'). Default is False
    """

    def __init__(self, user_groups: Dict[str, float],  user_profiles: Union[list, Ratings], original_ratings: Ratings,
                 out_dir: str = '.', file_name: str = 'pop_ratio_profile_vs_recs', pop_percentage: float = 0.2,
                 store_frame: bool = False, format: str = 'png', overwrite: bool = False):

        PlotMetric.__init__(self, out_dir, file_name, format, overwrite)
        GroupFairnessMetric.__init__(self, user_groups)

        if not 0 < pop_percentage <= 1:
            raise ValueError('Incorrect percentage! Valid percentage range: 0 < percentage <= 1')

        self._pop_by_item = get_item_popularity(original_ratings)

        if not isinstance(user_profiles, list):
            user_profiles = [user_profiles]

        self._user_profiles = user_profiles
        self.__pop_percentage = pop_percentage
        self.__user_groups = user_groups
        self.__store_frame = store_frame

    def perform(self, split: Split) -> pd.DataFrame:

        # in order to point to the right `user_profile` set each time the
        # `perform()` method is called, we pop the list but add the `user_profile` set
        # back at the end so that PopRatioProfileVsRecs is ready for another evaluation without
        # need to instantiate it again
        split_user_profile = self._user_profiles.pop(0)
        self._user_profiles.append(split_user_profile)

        predictions = split.pred

        most_pop_items = get_most_popular_items(self._pop_by_item, self.__pop_percentage)
        splitted_user_groups = self.split_user_in_groups(score_frame=split_user_profile, groups=self.user_groups,
                                                         pop_items=most_pop_items)

        split_result = {'user_group': [], 'profile_pop_ratio': [], 'recs_pop_ratio': []}
        data_to_plot = []
        labels = []
        for group_name in splitted_user_groups:

            # we don't consider users of the group for which we do not have any recommendation
            valid_group = splitted_user_groups[group_name].intersection(predictions.user_id_column)

            if len(valid_group) == 0:
                logger.warning(f"Group {group_name} won't be considered in the DeltaGap since no recs is available "
                               f"for any user of said group!")
                continue

            profile_group_ratings = split_user_profile.filter_ratings(user_list=valid_group)
            pred_group_recommendations = predictions.filter_ratings(user_list=valid_group)

            profile_pop_ratios = pop_ratio_by_user(profile_group_ratings, most_pop_items)
            recs_pop_ratios = pop_ratio_by_user(pred_group_recommendations, most_pop_items)

            profile_pop_ratios = list(profile_pop_ratios.values())
            recs_pop_ratios = list(recs_pop_ratios.values())

            split_result['user_group'].append(group_name)
            split_result['profile_pop_ratio'].append(profile_pop_ratios)
            split_result['recs_pop_ratio'].append(recs_pop_ratios)

            profile_data = profile_pop_ratios
            data_to_plot.append(profile_data)
            recs_data = recs_pop_ratios
            data_to_plot.append(recs_data)

            labels.append('{}\ngroup'.format(group_name))

        # agg backend is used to create plot as a .png file
        mpl.use('agg')

        # Create a figure instance
        fig = plt.figure()

        # Create an axes instance
        ax = fig.add_subplot()

        ax.set(title='Popularity ratio Profile vs Recs')

        # add patch_artist=True option to ax.boxplot()
        # to get fill color
        bp = ax.boxplot(np.array(data_to_plot, dtype=object), patch_artist=True)

        # make max y value always visible in the plot
        ax.set_ylim([0, 1])

        first_color = '#7570b3'
        second_color = '#b2df8a'
        fill_color_profile = '#004e98'
        fill_color_recs = '#ff6700'

        # change outline color, fill color and linewidth of the boxes
        for i, box in enumerate(bp['boxes']):
            # change outline color
            box.set(color=first_color, linewidth=2)
            # change fill color
            if i % 2 == 0:
                box.set(facecolor=fill_color_profile)
            else:
                box.set(facecolor=fill_color_recs)

        # change color and linewidth of the whiskers
        for whisker in bp['whiskers']:
            whisker.set(color=first_color, linewidth=2)

        # change color and linewidth of the caps
        for cap in bp['caps']:
            cap.set(color=first_color, linewidth=2)

        # change color and linewidth of the medians
        for median in bp['medians']:
            median.set(color=second_color, linewidth=2)

        # change the style of fliers and their fill
        for flier in bp['fliers']:
            flier.set(marker='o', color='#e7298a', alpha=0.5)

        # x ticks minor contains the vertical lines for better separate the various groups
        # a vertical line is 0.5 at the left of the first profile boxplot and another vertical line is at 0.5 of the
        # second boxplot for each group
        xticks_minor_tuples = [(i - 0.5, i + 1 + 0.5) for i in range(1, len(bp['boxes']) + 1, 2)]
        xticks_minor = list(itertools.chain.from_iterable(xticks_minor_tuples))

        # x ticks contains the "middle point" between the profile boxplot and recs boxplot
        # for each group
        x_ticks = [(i + (i+1)) / 2 for i in range(1, len(bp['boxes']) + 1, 2)]

        ax.set_xticks(x_ticks)
        ax.set_xticks(xticks_minor, minor=True)

        ax.set_xticklabels(labels)

        # make x_ticks_minor bigger, they are basically the vertical lines
        ax.tick_params(axis='x', which='minor', direction='out', length=25)
        # remove the tick and show only the label for the main ticks
        ax.tick_params(axis='x', which='major', length=0)

        # Remove top axes and right axes ticks
        ax.get_xaxis().tick_bottom()
        ax.get_yaxis().tick_left()

        # create the legend
        profile_patch = mpatches.Patch(color=fill_color_profile, label='Profile popularity')
        recs_patch = mpatches.Patch(color=fill_color_recs, label='Recs popularity')

        ax.legend(handles=[profile_patch, recs_patch], loc='upper center', bbox_to_anchor=(0.5, -0.15))

        plt.tight_layout()

        file_name = self.file_name
        self.save_figure(fig, file_name=file_name)

        score_frame = pd.DataFrame(split_result)
        if self.__store_frame:
            file_name = get_valid_filename(self.output_directory, file_name, 'csv', self.overwrite)
            score_frame.to_csv(os.path.join(self.output_directory, file_name), index=False)

        return pd.DataFrame(columns=['user_id', 'item_id', 'score'])

    def __str__(self):
        return "PopRatioProfileVsRecs"

    def __repr__(self):
        return f'PopRatioProfileVsRecs('\
               f'user_groups={self.__user_groups}, ' \
               f'out_dir={self.output_directory}, ' \
               f'file_name={self.file_name}, ' \
               f'pop_percentage={self.__pop_percentage}, ' \
               f'store_frame={self.__store_frame}, ' \
               f'format={self.format}, ' \
               f'overwrite={self.overwrite})'


class PopRecsCorrelation(PlotMetric):
    """
    This metric generates a plot which has as the X-axis the popularity of each item and as Y-axis the recommendation
    frequency, so that it can be easily seen the correlation between popular (niche) items and how many times are being
    recommended

    The popularity of an item is defined as the number of times it is rated in the `original_ratings` parameter
    divided by the total number of users in the `original_ratings`.

    The plot file will be saved as `out_dir/file_name.format`

    Since multiple split could be evaluated at once, the *overwrite* parameter comes into play:
    if is set to False, file with the same name will be saved as `file_name (1).format`, `file_name (2).format`, etc.
    so that for every split a plot is generated without overwriting any file previously generated

    There exists cases in which some items are not recommended even once, so in the graph could appear
    **zero recommendations**. One could change this behaviour thanks to the 'mode' parameter:

    - **mode='both'**: two graphs will be created, the first one containing eventual *zero recommendations*, the
      second one where *zero recommendations* are excluded. This additional graph will be stored as
      *out_dir/file_name_no_zeros.format* (the string '_no_zeros' will be added to the file_name chosen automatically)

    - **mode='w_zeros'**: only a graph containing eventual *zero recommendations* will be created

    - **mode='no_zeros'**: only a graph excluding eventual *zero recommendations* will be created. The graph will be
      saved as *out_dir/file_name_no_zeros.format* (the string '_no_zeros' will be added to the file_name chosen
      automatically)


    Args:
        original_ratings: `Ratings` object containing original interactions of the dataset that will be used to
            compute the popularity of each item (i.e. the number of times it is rated divided by the total number of
            users)
        out_dir (str): Directory where the plot will be saved. Default is '.', meaning that the plot will be saved
            in the same directory where the python script it's being executed
        file_name (str): Name of the plot file. Default is 'pop_recs_correlation'
        mode (str): Parameter which dictates which graph must be created. By default is 'both', so the graph with
            eventual zero recommendations as well as the graph excluding eventual zero recommendations will be created.
            Check the class documentation for more
        format (str): Format of the plot file. Could be 'jpg', 'svg', 'png'. Default is 'png'
        overwrite (bool): parameter which specifies if the plot saved must overwrite any file that as the same name
            ('file_name.format'). Default is False
    """

    def __init__(self, original_ratings: Ratings,
                 out_dir: str = '.',
                 file_name: str = 'pop_recs_correlation',
                 mode: str = 'both',
                 format: str = 'png', overwrite: bool = False):

        valid = {'both', 'no_zeros', 'w_zeros'}
        self.__mode = mode.lower()

        if self.__mode not in valid:
            raise ValueError("Mode {} is not supported! Modes available:\n"
                             "{}".format(mode, valid))

        self._pop_by_item = get_item_popularity(original_ratings)

        super().__init__(out_dir, file_name, format, overwrite)

    def __str__(self):
        return "PopRecsCorrelation"

    def __repr__(self):
        return f'PopRecsCorrelation('\
               f'out_dir={self.output_directory}, ' \
               f'file_name={self.file_name}, ' \
               f'mode={self.__mode}, ' \
               f'format={self.format}, ' \
               f'overwrite={self.overwrite})'

    def build_plot(self, x: list, y: list, title: str) -> matplotlib.figure.Figure:
        """
        Method which builds a matplotlib plot given x-axis values, y-axis values and the title of the plot.
        X-axis label and Y-axis label are hard-coded as 'Popularity' and 'Recommendation frequency' respectively.

        Args:
            x (list): List containing x-axis values
            y (list): List containing y-axis values
            title (str): title of the plot

        Returns:
            The matplotlib figure
        """
        fig = plt.figure()
        ax = fig.add_subplot()

        ax.set(xlabel='Popularity Ratio', ylabel='Recommendation frequency',
               title=title)

        ax.scatter(x, y, marker='o', s=20, c='orange', edgecolors='black',
                   linewidths=0.05)

        # automatic ticks but only integer ones
        ax.yaxis.set_major_locator(plticker.MaxNLocator(integer=True))

        return fig

    def build_w_zeros_plot(self, popularity: list, recommendations: list):
        """
        Method which builds and saves the plot containing eventual *zero recommendations*
        It saves the plot as *out_dir/filename.format*, according to their value passed in the constructor

        Args:
            popularity (list): x-axis values representing popularity of every item
            recommendations (list): y-axis values representing number of times every item has been recommended
        """
        title = 'Popularity Ratio - Recommendations Correlation'
        fig = self.build_plot(popularity, recommendations, title)

        file_name = self.file_name

        self.save_figure(fig, file_name)

    def build_no_zeros_plot(self, popularity: list, recommendations: list):
        """
        Method which builds and saves the plot **excluding** eventual *zero recommendations*
        It saves the plot as *out_dir/filename_no_zeros.format*, according to their value passed in the constructor.
        Note that the '_no_zeros' string is automatically added to the file_name chosen

        Args:
            popularity (list): x-axis values representing popularity of every item
            recommendations (list): y-axis values representing number of times every item has been recommended
        """
        title = 'Popularity Ratio - Recommendations Correlation (No zeros)'
        fig = self.build_plot(popularity, recommendations, title)

        file_name = self.file_name + '_no_zeros'

        self.save_figure(fig, file_name)

    def perform(self, split: Split):
        predictions = split.pred

        # Calculating num of recommendations for each item
        recs_by_item = Counter(predictions.item_id_column)
        popularities = list()
        recommendations = list()
        popularities_no_zeros = list()
        recommendations_no_zeros = list()

        at_least_one_zero = False
        for item, pop in self._pop_by_item.items():
            num_of_recs = recs_by_item[item]

            popularities.append(pop)
            recommendations.append(num_of_recs)

            if num_of_recs != 0:
                popularities_no_zeros.append(pop)
                recommendations_no_zeros.append(num_of_recs)
            else:
                at_least_one_zero = True

        # Both when possible
        if self.__mode == 'both':
            self.build_w_zeros_plot(popularities, recommendations)
            if at_least_one_zero:
                self.build_no_zeros_plot(popularities_no_zeros, recommendations_no_zeros)
            else:
                self.build_no_zeros_plot(popularities, recommendations)
                logger.warning("There's no zero recommendation!\n"
                               "The graph with 'no-zero' is identical to the 'w-zero' one!")

        elif self.__mode == 'w_zeros':
            self.build_w_zeros_plot(popularities, recommendations)

        elif self.__mode == 'no_zeros':
            self.build_no_zeros_plot(popularities_no_zeros, recommendations_no_zeros)

        return pd.DataFrame(columns=['user_id', 'item_id', 'score'])



from unittest import TestCase
import pandas as pd

from clayrs.content_analyzer import Ratings
from clayrs.recsys.graph_based_algorithm.graph_based_algorithm import GraphBasedAlgorithm
from clayrs.recsys.graphs.graph import ItemNode, UserNode, PropertyNode
from clayrs.recsys.graphs.nx_implementation.nx_full_graphs import NXFullGraph

from clayrs.recsys.graph_based_algorithm.page_rank.nx_page_rank import NXPageRank


class TestGraphBasedAlgorithm(TestCase):

    def setUp(self) -> None:
        ratings = pd.DataFrame.from_records([
            ("A000", "tt0114576", 1, "54654675"),
            ("A000", "tt0112453", -0.2, "54654675"),
            ("A001", "tt0114576", 0.8, "54654675"),
            ("A001", "tt0112896", -0.4, "54654675"),
            ("A000", "tt0113041", 0.6, "54654675"),
            ("A002", "tt0112453", -0.2, "54654675"),
            ("A002", "tt0113497", 0.5, "54654675"),
            ("A003", "tt0112453", -0.8, "54654675")],
            columns=["from_id", "to_id", "score", "timestamp"])
        ratings = Ratings.from_dataframe(ratings)

        self.graph = NXFullGraph(ratings)

        # GraphBasedAlgorithm is an abstract class, so we instantiate a subclass in order to test its methods
        self.alg = NXPageRank()

    def test_filter_result(self):
        rank = {UserNode("A000"): 0.5, ItemNode("tt0114576"): 0.5, UserNode("A001"): 0.5, ItemNode("tt0113497"): 0.5,
                ItemNode("tt0112453"): 0.5, PropertyNode("Nolan"): 0.5}

        # filter list with item i1, in this case graph parameter and user node parameter won't do anything
        result = self.alg.filter_result(graph=self.graph, result=rank, filter_list=[ItemNode('tt0114576')],
                                        user_node=UserNode("A000"))
        expected = {ItemNode("tt0114576"): 0.5}
        self.assertEqual(expected, result)

        # filter list with item i1 and item i2, in this case graph parameter and user node parameter won't do anything
        result = self.alg.filter_result(graph=self.graph, result=rank, filter_list=[ItemNode('tt0114576'),
                                                                                    PropertyNode('Nolan')],
                                        user_node=UserNode("A000"))
        expected = {ItemNode('tt0114576'): 0.5, PropertyNode("Nolan"): 0.5}
        self.assertEqual(expected, result)

        # filter with non existent nodes, result will be empty
        # in this case graph parameter and user node parameter won't do anything
        result = self.alg.filter_result(graph=self.graph, result=rank, filter_list=[ItemNode('non_existent')],
                                        user_node=UserNode("A000"))
        expected = {}
        self.assertEqual(expected, result)

        # clean result for user A000, the cleaned result will have only item nodes
        result = self.alg.filter_result(graph=self.graph, result=rank, filter_list=None,
                                        user_node=UserNode("A000"))
        expected = {ItemNode("tt0113497"): 0.5}

        self.assertEqual(expected, result)



import inspect
from abc import ABC, abstractmethod
from typing import Union, Any

import numpy as np
from scipy import sparse
from sklearn.ensemble import RandomForestClassifier
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier

from clayrs.utils.automatic_methods import autorepr


class Classifier(ABC):
    """
    Abstract class for Classifiers
    """

    def __init__(self, classifier, currentframe):
        self.__classifier = classifier

        self._repr_string = autorepr(self, currentframe)

    @property
    def classifier(self):
        return self.__classifier

    def fit(self, X: Union[np.ndarray, sparse.csr_matrix], Y: list = None):
        """
        Fit the classifier.

        Training data (X) is in the form:

            `X = [ (merged) representation for item 1, (merged) representation for item 2, ...]`

        where every item is a representation for an item (can be a merged one in case multiple representations
        are chosen)

        Target data (Y) is in the form:
            Y = [0, 1, ... ]
        where 0 represent a negative item, 1 represent a positive item

        Args:
            X: list containing Training data.
            Y: list containing Training targets.
        """
        self.classifier.fit(X, Y)

    def predict_proba(self, X_pred: list):
        """
        Predicts the probability for every item in X_pred.
        It uses the method predict_proba() from sklearn of the instantiated classifier

        It's in the form:
            X_pred = [ (merged) representation for item 1, (merged) representation for item 2, ...]
        where every item is a representation for an item (can be a merged one in case multiple representations
        are chosen)

        Args:
            X_pred: list containing data to predict.
        """
        return self.classifier.predict_proba(X_pred)

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        return self._repr_string


class SkSVC(Classifier):
    """
    Class that implements the SVC Classifier from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the classifier SVC directly from sklearn.

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html)

    The only parameter from sklearn that cannot be passed is the 'probability' parameter:
    it is set to True and cannot be changed
    """

    def __init__(self,
                 *,
                 C: Any = 1.0,
                 kernel: Any = "rbf",
                 degree: Any = 3,
                 gamma: Any = "scale",
                 coef0: Any = 0.0,
                 shrinking: Any = True,
                 tol: Any = 1e-3,
                 cache_size: Any = 200,
                 class_weight: Any = None,
                 verbose: Any = False,
                 max_iter: Any = -1,
                 decision_function_shape: Any = "ovr",
                 break_ties: Any = False,
                 random_state: Any = None):

        # Force the probability parameter at True, otherwise SVC won't predict_proba
        clf = SVC(C=C, kernel=kernel, degree=degree, gamma=gamma, coef0=coef0, shrinking=shrinking, tol=tol,
                  cache_size=cache_size, class_weight=class_weight, verbose=verbose, max_iter=max_iter,
                  decision_function_shape=decision_function_shape, break_ties=break_ties, random_state=random_state,
                  probability=True)

        super().__init__(clf, inspect.currentframe())

    def __str__(self):
        return "SkSVC"


class SkKNN(Classifier):
    """
    Class that implements the KNN Classifier from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the classifier KNN directly from sklearn.

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsClassifier.html)

    Since KNN implementation of sklearn has `n_neighbors = 5` as default, it can throw an exception if less sample in
    the training data are provided, so we change dynamically the n_neighbors parameter
    according to the number of samples if the dataset is too small and if no manual `n_neighbors` is set
    """

    def __init__(self, n_neighbors: Any = 5,
                 *,
                 weights: Any = "uniform",
                 algorithm: Any = "auto",
                 leaf_size: Any = 30,
                 p: Any = 2,
                 metric: Any = "minkowski",
                 metric_params: Any = None,
                 n_jobs: Any = None):
        clf = KNeighborsClassifier(n_neighbors=n_neighbors, weights=weights, algorithm=algorithm, leaf_size=leaf_size,
                                   p=p, metric=metric, metric_params=metric_params, n_jobs=n_jobs)

        super().__init__(clf, inspect.currentframe())

    def fit(self, X: Union[np.ndarray, sparse.csr_matrix], Y: list = None):
        # If the user did not pass a custom n_neighbor the algorithm tries to fix the error
        # if n_samples < n_neighbors
        if self.classifier.n_neighbors == 5 and X.shape[0] < self.classifier.n_neighbors:
            self.classifier.n_neighbors = X.shape[0]

        super().fit(X, Y)

    def __str__(self):
        return "SkKNN"


class SkRandomForest(Classifier):
    """
    Class that implements the Random Forest Classifier from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the classifier directly from sklearn

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html)

    """

    def __init__(self, n_estimators: Any = 100,
                 *,
                 criterion: Any = "gini",
                 max_depth: Any = None,
                 min_samples_split: Any = 2,
                 min_samples_leaf: Any = 1,
                 min_weight_fraction_leaf: Any = 0.0,
                 max_features: Any = "auto",
                 max_leaf_nodes: Any = None,
                 min_impurity_decrease: Any = 0.0,
                 bootstrap: Any = True,
                 oob_score: Any = False,
                 n_jobs: Any = None,
                 random_state: Any = None,
                 verbose: Any = 0,
                 warm_start: Any = False,
                 class_weight: Any = None,
                 ccp_alpha: Any = 0.0,
                 max_samples: Any = None):
        clf = RandomForestClassifier(n_estimators=n_estimators, criterion=criterion, max_depth=max_depth,
                                     min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf,
                                     min_weight_fraction_leaf=min_weight_fraction_leaf, max_features=max_features,
                                     max_leaf_nodes=max_leaf_nodes, min_impurity_decrease=min_impurity_decrease,
                                     bootstrap=bootstrap, oob_score=oob_score, n_jobs=n_jobs, random_state=random_state,
                                     verbose=verbose, warm_start=warm_start, class_weight=class_weight,
                                     ccp_alpha=ccp_alpha, max_samples=max_samples)

        super().__init__(clf, inspect.currentframe())

    def __str__(self):
        return "SkRandomForest"


class SkLogisticRegression(Classifier):
    """
    Class that implements the Logistic Regression Classifier from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the classifier directly from sklearn

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html)
    """

    def __init__(self, penalty: Any = "l2",
                 *,
                 dual: Any = False,
                 tol: Any = 1e-4,
                 C: Any = 1.0,
                 fit_intercept: Any = True,
                 intercept_scaling: Any = 1,
                 class_weight: Any = None,
                 random_state: Any = None,
                 solver: Any = "lbfgs",
                 max_iter: Any = 100,
                 multi_class: Any = "auto",
                 verbose: Any = 0,
                 warm_start: Any = False,
                 n_jobs: Any = None,
                 l1_ratio: Any = None):
        clf = LogisticRegression(penalty=penalty, dual=dual, tol=tol, C=C, fit_intercept=fit_intercept,
                                 intercept_scaling=intercept_scaling, class_weight=class_weight,
                                 random_state=random_state, solver=solver, max_iter=max_iter,
                                 multi_class=multi_class, verbose=verbose, warm_start=warm_start,
                                 n_jobs=n_jobs, l1_ratio=l1_ratio)

        super().__init__(clf, inspect.currentframe())

    def __str__(self):
        return "SkLogisticRegression"


class SkDecisionTree(Classifier):
    """
    Class that implements the Decision Tree Classifier from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the classifier directly from sklearn

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html)
    """

    def __init__(self, *,
                 criterion: Any = "gini",
                 splitter: Any = "best",
                 max_depth: Any = None,
                 min_samples_split: Any = 2,
                 min_samples_leaf: Any = 1,
                 min_weight_fraction_leaf: Any = 0.0,
                 max_features: Any = None,
                 random_state: Any = None,
                 max_leaf_nodes: Any = None,
                 min_impurity_decrease: Any = 0.0,
                 class_weight: Any = None,
                 ccp_alpha: Any = 0.0):
        clf = DecisionTreeClassifier(criterion=criterion, splitter=splitter, max_depth=max_depth,
                                     min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf,
                                     min_weight_fraction_leaf=min_weight_fraction_leaf, max_features=max_features,
                                     random_state=random_state, max_leaf_nodes=max_leaf_nodes,
                                     min_impurity_decrease=min_impurity_decrease, class_weight=class_weight,
                                     ccp_alpha=ccp_alpha)

        super().__init__(clf, inspect.currentframe())

    def __str__(self):
        return "SkDecisionTree"


class SkGaussianProcess(Classifier):
    """
    Class that implements the Gaussian Process Classifier from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the classifier directly from sklearn

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.GaussianProcessClassifier.html)
    """

    def __init__(self, kernel: Any = None,
                 *,
                 optimizer: Any = "fmin_l_bfgs_b",
                 n_restarts_optimizer: Any = 0,
                 max_iter_predict: Any = 100,
                 warm_start: Any = False,
                 copy_X_train: Any = True,
                 random_state: Any = None,
                 multi_class: Any = "one_vs_rest",
                 n_jobs: Any = None):

        clf = GaussianProcessClassifier(kernel=kernel, optimizer=optimizer, n_restarts_optimizer=n_restarts_optimizer,
                                        max_iter_predict=max_iter_predict, warm_start=warm_start,
                                        copy_X_train=copy_X_train, random_state=random_state,
                                        multi_class=multi_class, n_jobs=n_jobs)

        super().__init__(clf, inspect.currentframe())

    def __str__(self):
        return "SkGaussianProcess"

    def fit(self, X: Union[np.ndarray, sparse.csr_matrix], Y: list = None):
        self.classifier.fit(X.toarray() if isinstance(X, sparse.csr_matrix) else X, Y)

    def predict_proba(self, X_pred: Union[np.ndarray, sparse.csr_matrix]):
        return self.classifier.predict_proba(X_pred.toarray() if isinstance(X_pred, sparse.csr_matrix) else X_pred)
